"""Numpy-based loss functans that can be used by environments or non-pytorch models."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils/01_loss_functions.ipynb.

# %% auto 0
__all__ = ['pinball_loss', 'quantile_loss']

# %% ../nbs/00_utils/01_loss_functions.ipynb 3
from typing import Union, Optional

import numpy as np
from .utils import Parameter, check_parameter_types

import torch
import torch.nn as nn
import torch.nn.functional as F

# %% ../nbs/00_utils/01_loss_functions.ipynb 4
def pinball_loss(
            Y_true: np.ndarray, 
            Y_pred: np.ndarray,
            underage_cost: Parameter | np.ndarray,
            overage_cost: Parameter | np.ndarray,
            ) -> np.ndarray: # returns the cost per observation

    """

    Pinball loss calculating the cost of underestimating and overestimating the target value
    based on specific underage and overage costs. Used to evaulate the Newsvendor cost.

    """

    if isinstance(underage_cost, Parameter):
        underage_cost = underage_cost.get_value()
    if isinstance(overage_cost, Parameter):
        overage_cost = overage_cost.get_value()

    check_parameter_types(Y_true, Y_pred, underage_cost, overage_cost)

    # assert shapes
    assert Y_true.shape == Y_pred.shape, f"y_true and y_pred must have the same shape, but got {Y_true.shape} and {Y_pred.shape}"

    
    loss = np.maximum(Y_true - Y_pred, 0) * underage_cost + np.maximum(Y_pred - Y_true, 0) * overage_cost

    return loss

# %% ../nbs/00_utils/01_loss_functions.ipynb 6
def quantile_loss(
                Y_true: np.ndarray,
                Y_pred: np.ndarray,
                quantile: Union[float, Parameter],
                ) -> np.ndarray: # returns the cost per observation

    """
    Similar evaluation function to the pinball loss, but with the quantile of range
    [0, 1] as a parameter instead of SKU-specific cost levels for underage and overage.

    """

    if isinstance(quantile, Parameter):
        quantile = quantile.get_value()

    check_parameter_types(Y_true, Y_pred, quantile)
    
    # assert shapes
    assert Y_true.shape == Y_pred.shape, f"y_true and y_pred must have the same shape, but got {Y_true.shape} and {Y_pred.shape}"
    
    loss = np.maximum((Y_true - Y_pred) * quantile, (Y_pred - Y_true) * (1 - quantile))

    return loss





"""Agents based on Sample Average Approximation (SAA) or weighted Sample Average Approximation (wSAA)"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/41_NV_agents/10_NV_saa_agents.ipynb.

# %% auto 0
__all__ = ['BaseSAAagent', 'NewsvendorSAAagent', 'BasewSAAagent', 'NewsvendorRFwSAAagent']

# %% ../../../nbs/41_NV_agents/10_NV_saa_agents.ipynb 3
import logging

from abc import ABC, abstractmethod
from typing import Union, Optional, List
import numpy as np
import joblib
import os

from ...envs.base import BaseEnvironment
from ..base import BaseAgent
from ...utils import MDPInfo
from ...obsprocessors import FlattenTimeDimNumpy

from sklearn.ensemble import RandomForestRegressor
from sklearn.utils.validation import check_array

# %% ../../../nbs/41_NV_agents/10_NV_saa_agents.ipynb 4
class BaseSAAagent(BaseAgent):

    """
    Base class for Sample Average Approximation Agents, implementing the main method
    to find the quntile of some (weighted) empirical distribution.
    """

    def __init__(self,
                 environment_info: MDPInfo,
                 obsprocessors: Optional[List[object]] = None,
                 agent_name: str | None = None,
                 ):

        super().__init__(environment_info = environment_info, obsprocessors = obsprocessors, agent_name = agent_name)

        # check if FlattenTimeDimNumpy in obsprocessors
        if not any(isinstance(p, FlattenTimeDimNumpy) for p in self.obsprocessors):
            self.add_obsprocessor(FlattenTimeDimNumpy(allow_2d=True))

    def find_weighted_quantiles(self, weights, weightPosIndices, sl, y):
        
        """
        Find the weighted quantile of a range of data y. 
        It assumes that all arrays are of shape (n_samples, n_outputs). Note that it
        has not been tested for n_outputs > 1.
        """

        # test shapes have lenght 2 with error
        assert len(y.shape) == 2, "y should be of shape (n_samples, n_outputs)"

        n_outputs = y.shape[1]

        yWeightPos = y[weightPosIndices]

        if self.print:
            print(yWeightPos)
        
        q = []

        if len(weights.shape) == 1:
            weights = weights.reshape(-1, 1)
        
        for i in range(n_outputs):
            
            indicesYSort = np.argsort(yWeightPos[:, i])
            
            ySorted = yWeightPos[indicesYSort, i]
            
            distributionFunction = np.cumsum(weights[indicesYSort, i])

            decisionIndex = np.where(distributionFunction >= sl)[0][0]
            
            q.append(ySorted[decisionIndex])

        q = np.array(q)
        
        return q
    
    def _validate_X_predict(self, X):
        
        """Validate X data before prediction"""

        X = check_array(X)

        n_features = X.shape[1]
        if self.n_features_ != n_features:
            raise ValueError("Number of features of the model must match the input. "
                             "Model n_features is %s and input n_features is %s "
                             % (self.n_features_, n_features))
        return X

# %% ../../../nbs/41_NV_agents/10_NV_saa_agents.ipynb 8
class NewsvendorSAAagent(BaseSAAagent):

    """
    Newsvendor agent that uses Sample Average Approximation to find the quantile of the empirical distribution

    """

    def __init__(self,
                environment_info: MDPInfo,
                cu: float | np.ndarray, # underage cost
                co: float | np.ndarray, # overage cost
                obsprocessors: list[object] | None = None,
                agent_name: str = "SAA",
                ):

            # if float, convert to array
            cu = self.convert_to_numpy_array(cu)
            co = self.convert_to_numpy_array(co)

            self.sl = cu / (cu + co)
            self.fitted = False

            super().__init__(environment_info = environment_info, obsprocessors = obsprocessors, agent_name = agent_name)

    def fit(self,
            X: np.ndarray, # features will be ignored
            Y: np.ndarray) -> None:

        """

        Fit the agent to the data. The agent will find the quantile of the empirical distribution of the data.

        """

        # # potential line:
        # X, y = self._validate_data(X, y, multi_output=True)

        weights = np.ones(Y.shape)/Y.shape[0]
        weightPosIndices = np.arange(Y.shape[0])
        
        self.quantiles = self.find_weighted_quantiles(weights, weightPosIndices, self.sl, Y)

        self.fitted = True

    def draw_action_(self, 
                    observation: np.ndarray) -> np.ndarray: #
        """

        Draw an action from the quantile of the empirical distribution.

        """


        if self.fitted == False:
            return np.array([0.0])

        return self.quantiles


    def save(self,
                path: str, # The directory where the file will be saved.
                overwrite: bool=True): # Allow overwriting; if False, a FileExistsError will be raised if the file exists.
        
        """
        Save the quantiles to a file in the specified directory.

        """

        if not self.fitted:
            raise ValueError("Agent has not been fitted yet")

        os.makedirs(path, exist_ok=True)
        
        full_path = os.path.join(path, "saa_quantiles.npy")
        
        if os.path.exists(full_path):
            if not overwrite:
                raise FileExistsError(f"The file {full_path} already exists and will not be overwritten.")
            else:
                logging.warning(f"Overwriting file {full_path}")
                
        np.save(full_path, self.quantiles)

    def load(self, path: str): # Only the path to the folder is needed, not the file itself

        """
        Load the quantiles from a file.
        

        """

        full_path = os.path.join(path, "saa_quantiles.npy")
        
        if not os.path.exists(full_path):
            raise FileNotFoundError(f"The file {full_path} does not exist.")
        
        try:
            self.quantiles = np.load(full_path)
            self.fitted = True  # Assuming that loading the quantiles means the agent is now 'fitted'
            logging.info(f"Quantiles loaded successfully from {full_path}")
        except Exception as e:
            raise ValueError(f"An error occurred while loading the file: {e}")

# %% ../../../nbs/41_NV_agents/10_NV_saa_agents.ipynb 15
class BasewSAAagent(BaseSAAagent):


    """
    Base class for weighted Sample Average Approximation (wSAA) Agents

    """

    def __init__(self,
                environment_info: MDPInfo,
                cu: float | np.ndarray,
                co: float | np.ndarray,
                obsprocessors: list[object] | None = None,
                agent_name: str = "wSAA",
                ):  #


        # if float, convert to array
        self.cu = np.array([cu]) if isinstance(cu, float) else cu
        self.co = np.array([co]) if isinstance(co, float) else co

        cu = self.convert_to_numpy_array(cu)
        co = self.convert_to_numpy_array(co)

        self.sl = cu / (cu + co)
        
        self.fitted = False

        super().__init__(environment_info = environment_info, obsprocessors = obsprocessors, agent_name = agent_name)

    def fit(self,
            X: np.ndarray,
            Y: np.ndarray): #

        """

        Fit the agent to the data. The function will call _get_fitted_model which will
        train a machine learning model to determine the sample weightes (e.g., kNN, DT, RF).

        """
        
        # # potential line:
        # X, y = self._validate_data(X, y, multi_output=True)

        # get FlattenTimeDimNumpy obsprocessor
        flatten_obsprocessor = [p for p in self.obsprocessors if isinstance(p, FlattenTimeDimNumpy)][0]

        X = flatten_obsprocessor(X)

        if len(Y.shape) == 2 and Y.shape[1] == 1:
            Y = Y.flatten() 

        self._get_fitted_model(X, Y)

        if Y.ndim == 1:
            Y = np.reshape(Y, (-1, 1))

        # Training data
        self.Y_ = Y
        self.X_ = X
        self.n_samples_ = Y.shape[0]

        # Determine output settings
        self.n_outputs_ = Y.shape[1]
        self.n_features_ = X.shape[1]

        self.fitted=True

    def draw_action_(self, 
                    observation: np.ndarray) -> np.ndarray: # 

        """

        Draw an action based on the fitted model (see predict method)

        """

        if self.fitted == False:
            return np.array([0.0])
        
        return self.predict(observation)
    
    @abstractmethod
    def _get_fitted_model(self, X, y):
        """Initialise the underlying model - depending on the underlying machine learning model"""

    @abstractmethod
    def _calc_weights(self, sample):
        """Calculate the sample weights - depending on the underlying machine learning model"""

    def predict(self, 
                X: np.ndarray
    ) -> np.ndarray: #
        """Predict value for X by finding the quantiles of the empirical distribution based
        on the sample weights predicted by the underlying machine learning model.
        """

        X = self._validate_X_predict(X)  

        if self.print:
            print("X: ", X)

        weightsDataList = [self._calc_weights(row) for row in X]

        if self.print:
            print("weightsDataList: ", weightsDataList)

        pred = [self.find_weighted_quantiles(weights, weightPosIndices, self.sl, self.Y_) 
                for weights, weightPosIndices in weightsDataList]


        pred = np.array(pred)   

        if self.print:
            print("Predicted quantiles: ", pred)

        return pred

    def save(self,
                path: str, # The directory where the file will be saved.
                overwrite: bool=True): # Allow overwriting; if False, a FileExistsError will be raised if the file exists.
        """
        Save the scikit-learn model to a file in the specified directory.

        """

        if not self.fitted:
            raise ValueError("Agent has not been fitted yet")

        if not hasattr(self, 'model_') or self.model_ is None:
            raise ValueError("Agent has no model to save.")

        # Create directory if it does not exist
        os.makedirs(path, exist_ok=True)
        
        # Construct the file path using os.path.join for better cross-platform compatibility
        full_path = os.path.join(path, "model.joblib")
        
        if os.path.exists(full_path):
            if not overwrite:
                raise FileExistsError(f"The file {full_path} already exists and will not be overwritten.")
            else:
                logging.warning(f"Overwriting file {full_path}")
        
        # Save the model using joblib
        joblib.dump(self.model_, full_path)

    def load(self, path: str): # Only the path to the folder is needed, not the file itself
        """
        Load the scikit-learn model from a file.

        """
        
        # Construct the file path
        full_path = os.path.join(path, "model.joblib")
        
        if not os.path.exists(full_path):
            raise FileNotFoundError(f"The file {full_path} does not exist.")
        
        try:
            # Load the model using joblib
            self.model_ = joblib.load(full_path)
            self.fitted = True  # Assuming that loading the model means the agent is now 'fitted'
            logging.info(f"Model loaded successfully from {full_path}")
        except Exception as e:
            raise ValueError(f"An error occurred while loading the model: {e}")

# %% ../../../nbs/41_NV_agents/10_NV_saa_agents.ipynb 24
class NewsvendorRFwSAAagent(BasewSAAagent):

    """

    Newsvendor agent that uses weighted Sample Average Approximation based on Random Forest

    """

    def __init__(self,
                environment_info: MDPInfo,
                cu: float | np.ndarray, # underage cost
                co: float | np.ndarray, # overage cost
                obsprocessors: list[object] | None = None, # List of obsprocessors to apply to the observation
                n_estimators: int = 100,# The number of trees in the forest.
                criterion: str = "squared_error", # Function to measure the quality of a split.
                max_depth: int | None = None, # Maximum depth of the tree; None means unlimited.
                min_samples_split: int = 2, # Minimum samples required to split a node.
                min_samples_leaf: int = 1, # Minimum samples required to be at a leaf node.
                min_weight_fraction_leaf: float = 0.0, # Minimum weighted fraction of the total weights at a leaf node.
                max_features: int | float | str | None = 1.0, # Number of features to consider when looking for the best split.
                max_leaf_nodes: int | None = None, # Maximum number of leaf nodes; None means unlimited.
                min_impurity_decrease: float = 0.0, # Minimum impurity decrease required to split a node.
                bootstrap: bool = True, # Whether to use bootstrap samples when building trees.
                oob_score: bool = False, # Whether to use out-of-bag samples to estimate R^2 on unseen data.
                n_jobs: int | None = None, # Number of jobs to run in parallel; None means 1.
                random_state: int | np.random.RandomState | None = None, # Controls randomness for bootstrapping and feature sampling.
                verbose: int = 0, # Controls the verbosity when fitting and predicting.
                warm_start: bool = False, # If True, reuse solution from previous fit and add more estimators.
                ccp_alpha: float = 0.0, # Complexity parameter for Minimal Cost-Complexity Pruning.
                max_samples: int | float | None = None, # Number of samples to draw when bootstrap is True.
                monotonic_cst: np.ndarray | None = None, # Monotonic constraints for features.
                agent_name: str = "wSAA", # Default wSAA, change if it is needed to differentiate among different ML models
                ):
        self.criterion = criterion
        self.n_estimators = n_estimators
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.warm_start = warm_start
        self.ccp_alpha = ccp_alpha
        self.max_samples = max_samples
        self.monotonic_cst = monotonic_cst
        self.weight_function = "w1"

        super().__init__(environment_info = environment_info, cu = cu, co = co, obsprocessors = obsprocessors, agent_name = agent_name)

    def _get_fitted_model(self,
                            X: np.ndarray,
                            Y: np.ndarray): #

        """

        Fit the underlying machine learning model using all X and Y data in the train set.

        """

        model = RandomForestRegressor(
            criterion=self.criterion,
            n_estimators=self.n_estimators,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            max_features=self.max_features,
            max_leaf_nodes=self.max_leaf_nodes,
            min_impurity_decrease=self.min_impurity_decrease,
            bootstrap=self.bootstrap,
            oob_score=self.oob_score,
            n_jobs=self.n_jobs,
            random_state=self.random_state,
            verbose=self.verbose,
            warm_start=self.warm_start,
            ccp_alpha=self.ccp_alpha,
            max_samples=self.max_samples,
            monotonic_cst = self.monotonic_cst
        )

        self.model_ = model.fit(X, Y)
        self.train_leaf_indices_ = model.apply(X)

    def _calc_weights(self, sample: np.ndarray) -> tuple[np.ndarray, np.ndarray]: #

        """
        Calculate the sample weights based on the Random Forest model.

        """

        sample_leaf_indices = self.model_.apply([sample])
        if self.weight_function == "w1":
            n = np.sum(sample_leaf_indices == self.train_leaf_indices_, axis=0)
            treeWeights = (sample_leaf_indices == self.train_leaf_indices_) / n
            weights = np.sum(treeWeights, axis=1) / self.n_estimators
        else:
            n = np.sum(sample_leaf_indices == self.train_leaf_indices_)
            treeWeights = (sample_leaf_indices == self.train_leaf_indices_) / n
            weights = np.sum(treeWeights, axis=1)
        
        weightPosIndex = np.where(weights > 0)[0]
        weightsPos = weights[weightPosIndex]

        return (weightsPos, weightPosIndex)



"""Class to load the Kaggle_M5 dataset and prepare it for the context of inventory management"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/80_datasets/kaggle_m5.ipynb.

# %% auto 0
__all__ = ['KaggleM5DatasetLoader']

# %% ../../nbs/80_datasets/kaggle_m5.ipynb 3
import logging
logging.basicConfig(level=logging.INFO)

import numpy as np
import os
import pandas as pd

# %% ../../nbs/80_datasets/kaggle_m5.ipynb 4
class KaggleM5DatasetLoader():

    """ Class to download the Kaggle M5 dataset and apply some preprocessing steps
    to prepare it for application in inventory management. """

    def __init__(self, data_path, overwrite=False, product_as_feature=False):
        self.create_paths(data_path)
        self.check_data_path(data_path, overwrite)
        self.product_as_feature = product_as_feature
    
    def load_dataset(self):

        """ Main function to load the dataset. """

        if self.download_data_flag:
            logging.info("Downloading dataset from Kaggle")
            self.download_data()
        else:
            logging.info("Using existing data from disk")
        logging.info("Importing data")
        self.import_from_folder()
        logging.info("Preprocessing data")
        self.preprocess_pipeline()

        return self.demand, self.SKU_features, self.time_features, self.time_SKU_features, self.mask

    def check_data_path(self, data_path, overwrite):

        """ Check if the data path exists and if the data should be downloaded. """

        if not os.path.exists(data_path):
            os.makedirs(data_path)
            self.download_data_flag = True
        else:
            if os.path.exists(self.calendar_path) and os.path.exists(self.sale_path) and os.path.exists(self.price_path):
                self.download_data_flag = True if overwrite else False
            else:
                self.download_data_flag = True

    def create_paths(self, data_path):

        """ Create the paths for the data files. """

        self.data_path = data_path
        self.calendar_path = os.path.join(data_path, "calendar.csv")
        self.sale_path = os.path.join(data_path, "sales_train_evaluation.csv")
        self.price_path = os.path.join(data_path, "sell_prices.csv")

    def preprocess_pipeline(self):
    
        """ Apply simple preprocessing steps to the data. """
        
        logging.info("--Creating catogory mapping and features")  
        self.sale["id"] = self.sale["id"].str.replace("_evaluation", "")
        unique_mapping = self.sale[['item_id', 'dept_id', 'cat_id', 'store_id']].drop_duplicates()
        unique_mapping["SKU_id"] = unique_mapping["item_id"] + "_" + unique_mapping["store_id"]
        unique_mapping["state"] = unique_mapping["store_id"].str.split("_", expand=True)[0]
        unique_mapping.set_index("SKU_id", inplace=True)

        dummy_columns = ["dept_id", "cat_id", "store_id", "state"]
        if self.product_as_feature:
            dummy_columns.append("item_id")
        categories = pd.get_dummies(unique_mapping[dummy_columns], drop_first=True) 

        logging.info("--Preparing sales time series data")
        id = self.sale["id"]
        self.sale = self.sale.iloc[:,6:]
        self.sale["SKU_id"] = id
        self.sale.set_index("SKU_id", inplace=True)
        self.sale = self.sale.transpose()
        self.sale.reset_index(inplace=True, drop=True)
        self.sale.rename_axis(None, axis=1, inplace=True)
        self.sale = self.sale[unique_mapping.index]
        
        logging.info("--Preparing calendric information")
        self.calendar = self.calendar[:1941] # we are only interested in the data we also have sales data for
        self.calendar.drop(["date", "d", "weekday"], axis=1, inplace=True)
        self.calendar["trend"] = np.arange(1, len(self.calendar)+1)
        # For even larger datasets this coule be done at runtime when creating samples, but for this one it should be fine memory-wise
        dummy_columns = ["wday", "month", "event_name_1", "event_type_1", "event_name_2", "event_type_2"]
        self.calendar = pd.get_dummies(self.calendar, columns=dummy_columns, drop_first=True)
        cols = self.calendar.columns.tolist()
        cols.insert(4, cols.pop(1))
        self.calendar = self.calendar[cols]

        logging.info("--Preparing snap features")
        snap_features = self.calendar[["snap_CA", "snap_TX", "snap_WI"]].copy()
        self.calendar.drop(["snap_CA", "snap_TX", "snap_WI"], axis=1, inplace=True)

        new_snap_features = pd.DataFrame(index=range(snap_features.shape[0]), columns=unique_mapping.index)
        for sku in unique_mapping.index:
            state_code = sku.split('_')[-2]
            snap_column = f"snap_{state_code}"
            new_snap_features[sku] = snap_features[snap_column].values
        snap_features = new_snap_features

        logging.info("--Preparing price information")
        self.price["SKU_id"] = self.price["item_id"] + "_" + self.price["store_id"]
        self.price.drop(["store_id", "item_id"], axis=1, inplace=True)
        self.price = self.price.pivot_table(index='wm_yr_wk', columns=['SKU_id'], values='sell_price')
        self.price = self.price[unique_mapping.index]

        logging.info("--Creating indicator table if products are available for purchase")
        self.available = self.price.copy()
        self.available = self.available.notnull().astype(int)
        # fill missing values for price (indicated in the available table)
        self.price.fillna(0, inplace=True)

        logging.info("--Preparing final outputs and ensure consistency of time and feature dimensions")
        wm_yr_wk_per_day = self.calendar[["wm_yr_wk"]]
        self.price = self.price.reset_index()
        missing_wm_yr_wk_in_price = wm_yr_wk_per_day[~wm_yr_wk_per_day["wm_yr_wk"].isin(self.price["wm_yr_wk"])]
        if not missing_wm_yr_wk_in_price.empty:
            raise ValueError("The following wm_yr_wk values are in calendar but not in price: ", missing_wm_yr_wk_in_price.tolist())
        self.price = self.price.merge(wm_yr_wk_per_day, on="wm_yr_wk", how="right")
        self.price.drop(["wm_yr_wk"], axis=1, inplace=True)
        self.available = self.available.reset_index()
        self.available = self.available.merge(wm_yr_wk_per_day, on="wm_yr_wk", how="right")
        self.available.drop(["wm_yr_wk"], axis=1, inplace=True)
        self.calendar.drop(["wm_yr_wk"], axis=1, inplace=True)

        price_multi_index = pd.MultiIndex.from_product([['Price'], self.price.columns], names=['Type', 'SKU'])
        self.price.columns = price_multi_index
        snap_multi_index = pd.MultiIndex.from_product([['Snap'], snap_features.columns], names=['DataType', 'SKU'])
        snap_features.columns = snap_multi_index
        time_SKU_features = pd.concat([self.price, snap_features], axis=1)
       
        self.demand = self.sale
        self.SKU_features = categories # features that are not time-dependent
        self.time_features = self.calendar # features that are time-dependent
        self.time_SKU_features = time_SKU_features # features taht are time- and SKU-dependent
        self.mask = self.available # A mask that can either mask datapoints during training or be used as a feature

    def import_from_folder(self):
        
        """ Import data from a folder. """

        self.calendar = pd.read_csv(self.calendar_path)
        self.sale = pd.read_csv(self.sale_path)
        self.price = pd.read_csv(self.price_path)
    
    def download_data(self):

        """ Download the data directly from Kaggle. """

        raise NotImplementedError(
            "Download function not implemented yet - please download the data manually\n"
            "from Kaggle and place it in the directory specified in the 'data_path' variable during initialization")



"""Models that approximate a function"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/60_approximators/21_critic_networks.ipynb.

# %% auto 0
__all__ = ['RNNWrapper', 'BaseApproximator', 'BaseApproximatorMLP', 'RNNMLPHybrid', 'BaseApproximatorRNN', 'MLPStateAction',
           'MLPState', 'MLPActor', 'RNNActor', 'RNNStateAction']

# %% ../nbs/60_approximators/21_critic_networks.ipynb 3
# import logging
# logging_level = logging.DEBUG

from abc import ABC, abstractmethod
from typing import Union, Tuple, List
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader

import time

# %% ../nbs/60_approximators/21_critic_networks.ipynb 4
class RNNWrapper(nn.Module):
    def __init__(self, rnn_cell_class, *args, **kwargs):
        """
        Initializes the RNNWrapper with the specified RNN cell.
        
        Parameters:
        - rnn_cell_class: The RNN cell class (e.g., nn.GRU, nn.LSTM, nn.RNN).
        - *args, **kwargs: The arguments and keyword arguments to be passed to the RNN cell.
        """
        super(RNNWrapper, self).__init__()
        self.rnn = rnn_cell_class(*args, **kwargs)

    def forward(self, x):
        output, _ = self.rnn(x)  # Extract and return only the output
        return output

    @classmethod
    def create(cls, rnn_cell_class):
        """
        A factory method to create a new RNNWrapper subclass with a specific RNN cell.
        
        Parameters:
        - rnn_cell_class: The RNN cell class to be wrapped (e.g., nn.GRU, nn.LSTM).
        
        Returns:
        - A new subclass of RNNWrapper.
        """
        class SpecificRNNWrapper(cls):
            def __init__(self, *args, **kwargs):
                super(SpecificRNNWrapper, self).__init__(rnn_cell_class, *args, **kwargs)

        return SpecificRNNWrapper

# %% ../nbs/60_approximators/21_critic_networks.ipynb 5
class BaseApproximator():

    """ Some basic functions for approximators """

    def __init__(self):
        super().__init__()

    def init_weights(self, layer, init_method, activation):
        """ Initialize the weights of a layer """
        init_method_function = self.select_init_method(init_method)
        
        # Check if initialization method requires gain
        if init_method in ["xavier_uniform", "xavier_normal"]:
            if activation == "identity":
                activation = "linear"
            gain = nn.init.calculate_gain(activation)
            init_method_function(layer.weight, gain=gain)
        else:
            init_method_function(layer.weight)
            
    def init_rnn_weights(self, rnn_layer, init_method):
        """Initialize the weights for the RNN layer."""
        init_method_function = self.select_init_method(init_method)

        for name, param in rnn_layer.named_parameters():
            if 'weight' in name:
                # Initialize weights using the selected method
                init_method_function(param)
            elif 'bias' in name:
                # Initialize biases to zero
                nn.init.constant_(param, 0)
                
    @staticmethod
    def select_rnn_cell(RNN_cell):
        """ Select the RNN cell based on input string """
        RNN_cell = RNN_cell.lower()  # Convert input to lowercase for consistency
        if RNN_cell == "gru":
            rnn_cell_class = nn.GRU
        elif RNN_cell == "lstm":
            rnn_cell_class = nn.LSTM
        elif RNN_cell == "rnn":
            rnn_cell_class = nn.RNN
        else:
            raise ValueError(f"RNN cell '{RNN_cell}' not recognized")

        return RNNWrapper.create(rnn_cell_class)

    @staticmethod
    def select_init_method(init_method):
        """ Select the initialization method based on input string """
        init_method = init_method.lower()
        if init_method in ["xavier_uniform", "xavier"]:
            return nn.init.xavier_uniform_
        elif init_method in ["xavier_normal", "xaviernorm"]:
            return nn.init.xavier_normal_
        elif init_method in ["he_uniform", "kaiming_uniform"]:
            return nn.init.kaiming_uniform_
        elif init_method in ["he_normal", "kaiming_normal"]:
            return nn.init.kaiming_normal_
        elif init_method in ["normal", "gaussian"]:
            return nn.init.normal_
        elif init_method == "uniform":
            return nn.init.uniform_
        else:
            raise ValueError("Initialization method not recognized")

    @staticmethod
    def select_activation(activation):
        """ Select the activation function based on input string """
        activation = activation.lower()  # Convert input to lowercase for consistency
        if activation == "relu":
            return nn.ReLU
        elif activation == "sigmoid":
            return nn.Sigmoid
        elif activation == "tanh":
            return nn.Tanh
        elif activation == "elu":
            return nn.ELU
        elif activation == "leakyrelu":
            return nn.LeakyReLU
        elif activation == "identity":
            return nn.Identity
        else:
            raise ValueError(f"Activation function {activation} not recognized")

    def forward(self, x):
        """ Forward pass through the network - overwrite this if necessary """
        return self.model(x)

# %% ../nbs/60_approximators/21_critic_networks.ipynb 6
class BaseApproximatorMLP(BaseApproximator, nn.Module):

    """ Some basic functions for approximators """

    def __init__(self):
        super().__init__()
    
    def build_MLP(  self,
                    input_size: int,
                    output_size: int,
                    hidden_layers: list,
                    activation: str = "relu",
                    drop_prob: float = 0.0,
                    batch_norm: bool = False,
                    final_activation: str = "identity",
                    init_method: str = "xavier_uniform" # Parameter for initialization
                  ):

        """ Builds a multi-layer perceptron (MLP) """

        HiddenActivation = self.select_activation(activation)
        FinalActivation = self.select_activation(final_activation)

        layers = []

        # Hidden layers
        last_size = input_size
        for num_neurons in hidden_layers:
            
            linear_layer = nn.Linear(last_size, num_neurons)
            self.init_weights(linear_layer, init_method, activation)
            layers.append(linear_layer)

            if batch_norm:
                layers.append(nn.BatchNorm1d(num_neurons))
        
            layers.append(HiddenActivation())
            layers.append(nn.Dropout(p=drop_prob))

            last_size = num_neurons
        
        # Output layer
        output_layer = nn.Linear(last_size, output_size)
        self.init_weights(output_layer, init_method, final_activation)
        layers.append(output_layer)
        layers.append(FinalActivation())

        # Combine layers
        model = nn.Sequential(*layers)

        self.model = model

# %% ../nbs/60_approximators/21_critic_networks.ipynb 7
class RNNMLPHybrid(nn.Module, BaseApproximator):

    """ A hybrid model combining an RNN and an MLP """

    def __init__(self, 
                 RNN_input_size: int,
                 MLP_input_size: int | None,
                 output_size: int,
                 num_hidden_units_RNN: int,
                 hidden_layers_RNN: int,
                 hidden_layers_MLP: List[int],
                 hidden_layers_input_MLP: List[int] | None,
                 RNN_cell: nn.Module,
                 activation: nn.Module,
                 final_activation: nn.Module,
                 drop_prob: float,
                 batch_norm: bool,
                 init_method: str):
        super(RNNMLPHybrid, self).__init__()

        HiddenActivation = self.select_activation(activation)
        FinalActivation = self.select_activation(final_activation)
        RNNCell = self.select_rnn_cell(RNN_cell)

        # RNN
        # RNN layers

        rnn_layers = []
        rnn = RNNCell(RNN_input_size, num_hidden_units_RNN, hidden_layers_RNN, batch_first=True, dropout=drop_prob)
        self.init_rnn_weights(rnn, init_method)
        hidden_activation_rnn = HiddenActivation()  # Activation used after RNN layers
        rnn_layers.append(rnn)
        rnn_layers.append(hidden_activation_rnn)

        self.rnn = nn.Sequential(*rnn_layers)

        # Input MLP, if required:
        last_size = 0 if MLP_input_size is None else MLP_input_size
        if hidden_layers_input_MLP is not None:

            if last_size == 0:
                raise ValueError("MLP input size must be specified if input MLP is used")
            
            layers_input_MLP = []
            for num_neurons in hidden_layers_input_MLP:
                linear_layer = nn.Linear(last_size, num_neurons)
                self.init_weights(linear_layer, init_method, activation)
                layers_input_MLP.append(linear_layer)
                if batch_norm:
                    layers_input_MLP.append(nn.BatchNorm1d(num_neurons))
                layers_input_MLP.append(HiddenActivation())
                layers_input_MLP.append(nn.Dropout(p=drop_prob))
                last_size = num_neurons
            
            self.input_mlp = nn.Sequential(*layers_input_MLP)
        else:
            self.input_mlp = None
        
        # Main MLP layers
        layers_MLP = []
        last_size = num_hidden_units_RNN + last_size
        for num_neurons in hidden_layers_MLP:
            linear_layer = nn.Linear(last_size, num_neurons)
            self.init_weights(linear_layer, init_method, activation)
            layers_MLP.append(linear_layer)
            if batch_norm:
                layers_MLP.append(nn.BatchNorm1d(num_neurons))
            layers_MLP.append(HiddenActivation())
            layers_MLP.append(nn.Dropout(p=drop_prob))
            last_size = num_neurons

        # Output layer
        output_layer = nn.Linear(last_size, output_size)
        self.init_weights(output_layer, init_method, final_activation)
        layers_MLP.append(output_layer)

        self.main_mlp = nn.Sequential(*layers_MLP)

    
    def forward(self, x_rnn, x_mlp=None):
        # RNN

        rnn_out = self.rnn(x_rnn) # Only one output due to the wrapper
        rnn_out = rnn_out[:, -1, :]  # Take the last output of the RNN
        
        # Input MLP
        if x_mlp is not None:
            if self.input_mlp is not  None:
                x_mlp = self.input_mlp(x_mlp)
            x = torch.cat((rnn_out, x_mlp), dim=1)
        else:
            x = rnn_out

        # Main MLP
        x = self.main_mlp(x)

        return x

# %% ../nbs/60_approximators/21_critic_networks.ipynb 8
class BaseApproximatorRNN(BaseApproximator, nn.Module):

    """ Some basic functions for approximators """

    def __init__(self):
        super().__init__()

    def build_RNN(  self,
                    input_size: int | List[int], # is List, it means that multiple inputs are used. The first element is alwas for the RNN, the rest for the MLP
                    output_size: int,
                    hidden_layers_RNN:int,
                    num_hidden_units_RNN: int,
                    hidden_layers_MLP:List,
                    hidden_layers_input_MLP: List | None = None, # If a separate MLP is used for (potential) MLP input
                    RNN_cell: str = "GRU",
                    activation: str = "relu",
                    drop_prob: float = 0.0,
                    batch_norm: bool = False,
                    final_activation: str = "identity",
                    init_method: str = "xavier_uniform" # Parameter for initialization
                  ):

        """ Builds a recurrent neural network (RNN) """

        if isinstance(input_size, int):
            RNN_input_size = input_size
            MLP_input_size = None
        elif isinstance(input_size, list):

            if len(input_size) != 2:
                raise ValueError(f"Input size must be a list of length 2 (got {len(input_size)}) with elementes (RNN_input_size, MLP_input_size)")

            RNN_input_size = input_size[0]
            MLP_input_size = input_size[1]

        else:
            raise ValueError("Input size must be an integer or a list of integers")
    
        self.model = RNNMLPHybrid(  RNN_input_size,
                                    MLP_input_size,
                                    output_size,
                                    num_hidden_units_RNN,
                                    hidden_layers_RNN,
                                    hidden_layers_MLP,
                                    hidden_layers_input_MLP,
                                    RNN_cell,
                                    activation,
                                    final_activation,
                                    drop_prob,
                                    batch_norm,
                                    init_method,
                                    )


        

# %% ../nbs/60_approximators/21_critic_networks.ipynb 9
class MLPStateAction(BaseApproximatorMLP):

    """Multilayer perceptron model for critic networks that take
    both states and actions as inputs to output the q-value"""

    def __init__(self,
                    input_shape: Tuple | List[Tuple], # number of features
                    output_shape: Tuple, # number of outputs/actions
                    hidden_layers: list, # list of number of neurons in each hidden layer
                    activation: str = "relu",
                    drop_prob: float = 0.0, # dropout probability
                    batch_norm: bool = False, # whether to apply batch normalization
                    final_activation: str = "identity", # whether to apply ReLU activation to the output
                    init_method: str = "xavier_uniform",  # Parameter for initialization
                    use_cuda: bool = False, # handled by mushroomRL, not used here
                    dropout: bool = False # legacy parameter to ensure compatibility, use drop_prob instead
                    ):

        super().__init__()

        # if input shape is list, then concatenate the elements
        if isinstance(input_shape, list):
            input_shape = (sum([shape[0] for shape in input_shape]),)
        
        self.build_MLP(    input_shape[0],
                            output_shape[0],
                            hidden_layers,
                            activation, 
                            drop_prob,
                            batch_norm,
                            final_activation,
                            init_method)

    def forward(self, state, action):


        state_action = torch.cat([state.float(), action.float()], dim=1)

        q = self.model(state_action)

        # TODO: check if squeeze is necessary
        # return q
        return torch.squeeze(q)

# %% ../nbs/60_approximators/21_critic_networks.ipynb 10
class MLPState(BaseApproximatorMLP):

    """Multilayer perceptron model for critic networks that take
    both states and actions as inputs to output the q-value"""

    def __init__(self,
                    input_shape: Tuple, # number of features
                    output_shape: Tuple, # number of outputs/actions
                    hidden_layers: list, # list of number of neurons in each hidden layer
                    activation: str = "relu",
                    drop_prob: float = 0.0, # dropout probability
                    batch_norm: bool = False, # whether to apply batch normalization
                    final_activation: str = "identity", # whether to apply ReLU activation to the output
                    init_method: str = "xavier_uniform",  # Parameter for initialization
                    use_cuda: bool = False, # handled by mushroomRL, not used here
                    dropout: bool = False # legacy parameter to ensure compatibility, use drop_prob instead
                    ):

        super().__init__()
        
        self.build_MLP(    input_shape[0],
                            output_shape[0],
                            hidden_layers,
                            activation, 
                            drop_prob,
                            batch_norm,
                            final_activation,
                            init_method)

    def forward(self, state):

        state = state.float()

        q = self.model(state)

        # TODO: check if squeeze is necessary
        # return q.squeeze()
        return q

# %% ../nbs/60_approximators/21_critic_networks.ipynb 11
class MLPActor(BaseApproximatorMLP):

    """Multilayer perceptron model for critic networks that take
    both states and actions as inputs to output the q-value"""

    def __init__(self,
                    input_shape: Tuple, # number of features
                    output_shape: Tuple, # number of outputs/actions
                    hidden_layers: list, # list of number of neurons in each hidden layer
                    activation: str = "relu",
                    drop_prob: float = 0.0, # dropout probability
                    batch_norm: bool = False, # whether to apply batch normalization
                    final_activation: str = "identity", # whether to apply ReLU activation to the output
                    init_method: str = "xavier_uniform",  # Parameter for initialization
                    use_cuda: bool = False,
                    dropout: bool = False, # legacy parameter to ensure compatibility, use drop_prob instead
                    **kwargs
                    ): 

        super().__init__()
        
        self.build_MLP(    input_shape[0],
                            output_shape[0],
                            hidden_layers,
                            activation, 
                            drop_prob,
                            batch_norm,
                            final_activation,
                            init_method)

    def forward(self, state):

        state = state.float()

        a = self.model(state)

        return a

# %% ../nbs/60_approximators/21_critic_networks.ipynb 12
class RNNActor(BaseApproximatorRNN):

    """Multilayer perceptron model for critic networks that take
    both states and actions as inputs to output the q-value"""

    def __init__(self,
                    input_shape: List[Tuple], # input shape, must be exaclty as input shape into agent for mushroom_rl to work
                    output_shape: Tuple, # number of outputs/actions
                    hidden_layers_RNN: int, # number of initial hidden RNN layers
                    num_hidden_units_RNN: int, # number of neurons in the RNN layers
                    hidden_layers_MLP: List, # list of number of neurons in each hidden MLP layer, following the RNN layers
                    hidden_layers_input_MLP: List | None = None, # If a separate MLP is used for (potential) MLP input
                    RNN_cell: str = "GRU", # RNN cell type
                    activation: str = "relu",
                    drop_prob: float = 0.0, # dropout probability
                    batch_norm: bool = False, # whether to apply batch normalization
                    final_activation: str = "identity", # whether to apply ReLU activation to the output
                    init_method: str = "xavier_uniform",  # Parameter for initialization
                    use_cuda: bool = False,
                    dropout: bool = False, # legacy parameter to ensure compatibility, use drop_prob instead
                    input_shape_: List[Tuple] = None, # input shape for composite spaces
                    **kwargs
                    ): 

        super().__init__()

        if input_shape_ is not None:
            input_shape = input_shape_

        if isinstance(input_shape, tuple):
            if len(input_shape) != 2:
                raise ValueError(f"Input shape must be a tuple with dimensions (time_steps, features), got {input_shape}")
            input_size = input_shape[1]
            self.rnn_shape = input_shape
            self.mlp_shape = None
            self.get_time_input = True
        else:
            if len(input_shape) > 2:
                raise ValueError(f"Input shape must be a tuple or a list of tuples with length 1 or 2, got length {len(input_shape)}")
            if len(input_shape) == 2:
                input_size = [input_shape[0][1], input_shape[1][0]]
                self.rnn_shape = input_shape[0]
                self.mlp_shape = input_shape[1]
            else:
                input_size = input_shape[0][1]
                self.rnn_shape = input_shape[0]
                self.mlp_shape = None
            self.get_time_input = False

        self.build_RNN(     input_size,
                            output_shape[0],
                            hidden_layers_RNN,
                            num_hidden_units_RNN,
                            hidden_layers_MLP,
                            hidden_layers_input_MLP,
                            RNN_cell,
                            activation, 
                            drop_prob,
                            batch_norm,
                            final_activation,
                            init_method)

    def forward(self, state):
        
        if self.get_time_input: # already in the right 2d (or 3d with batch) format
            a = self.model(state.float(), None)
        
        else:
            if state.dim() == 1:
                a = self.forward_without_batch(state.float())
            else:
                a = self.forward_with_batch(state.float())
                
        return a

    def forward_with_batch(self, state):
        if self.mlp_shape is not None:
            rnn_input = state[:, :self.rnn_shape[0] * self.rnn_shape[1]]
            mlp_input = state[:, self.rnn_shape[0] * self.rnn_shape[1]:]
        else:
            rnn_input = state
            mlp_input = None
    
        # Reshape rnn_input to (batch_size, time, features)
        rnn_input = rnn_input.view(-1, self.rnn_shape[0], self.rnn_shape[1])
        
        return self.model(rnn_input, mlp_input)

    def forward_without_batch(self, state):
        if self.mlp_shape is not None:
            rnn_input = state[:self.rnn_shape[0] * self.rnn_shape[1]]
            mlp_input = state[self.rnn_shape[0] * self.rnn_shape[1]:]
        else:
            rnn_input = state
            mlp_input = None
        
        # Reshape rnn_input to (time, features)
        rnn_input = rnn_input.view(self.rnn_shape[0], self.rnn_shape[1])
        
        return self.model(rnn_input, mlp_input)

# %% ../nbs/60_approximators/21_critic_networks.ipynb 13
class RNNStateAction(BaseApproximatorRNN):

    """Multilayer perceptron model for critic networks that take
    both states and actions as inputs to output the q-value"""

    def __init__(self,
                    input_shape: List[Tuple], # input shape, must be exaclty as input shape into agent for mushroom_rl to work
                    output_shape: Tuple, # Output shape
                    hidden_layers_RNN: int, # number of initial hidden RNN layers
                    num_hidden_units_RNN: int, # number of neurons in the RNN layers
                    hidden_layers_MLP: List, # list of number of neurons in each hidden MLP layer, following the RNN layers
                    hidden_layers_input_MLP: List | None = None, # structure of MLP to speratly process non-RNN input
                    RNN_cell: str = "GRU", # RNN cell type
                    activation: str = "relu",
                    drop_prob: float = 0.0, # dropout probability
                    batch_norm: bool = False, # whether to apply batch normalization
                    final_activation: str = "identity", # whether to apply ReLU activation to the output
                    init_method: str = "xavier_uniform",  # Parameter for initialization
                    use_cuda: bool = False,
                    dropout: bool = False, # legacy parameter to ensure compatibility, use drop_prob instead
                    input_shape_: List[Tuple] = None, # input shape for composite spaces
                    **kwargs
                    ): 

        super().__init__()

        if input_shape_ is not None:
            input_shape = input_shape_

        # check that input lenght of list is 2 or 3
        if len(input_shape) != 2:
            raise ValueError(f"Input shape must be a list of length 2, got {len(input_shape)}")
        
        action_input = input_shape[1][0]
    
        if isinstance(input_shape[0], tuple):
            if len(input_shape) != 2:
                raise ValueError(f"Input shape must be a tuple with dimensions (time_steps, features), got {input_shape}")
            self.rnn_shape = input_shape[0]
            non_time_features = 0
            self.get_time_input = True
        else:
            if isinstance(input_shape[0], list):
                if len(input_shape[0]) > 2:
                    raise ValueError(f"Input shape must be a list of length 1 or 2, got {len(input_shape[0])}")
                self.rnn_shape = input_shape[0][0]
                if len(input_shape[0]) == 2:
                    non_time_features = input_shape[0][1][0]
                else:
                    non_time_features = 0
                self.get_time_input = False
            else:
                raise ValueError("Input shape for composite spaces must be a list")
        
        self.mlp_shape = (action_input + non_time_features,)

        self.build_RNN(     [self.rnn_shape[1], self.mlp_shape[0]],
                            output_shape[0],
                            hidden_layers_RNN,
                            num_hidden_units_RNN,
                            hidden_layers_MLP,
                            hidden_layers_input_MLP,
                            RNN_cell,
                            activation, 
                            drop_prob,
                            batch_norm,
                            final_activation,
                            init_method)

    def forward(self, state, action):

        if self.get_time_input: # already in the right 2d (or 3d with batch) format
            q = self.model(state.float(), action.float())
        
        else:
            if state.dim() == 1:
                q = self.forward_without_batch(state.float(), action.float())
            else:
                q = self.forward_with_batch(state.float(), action.float())

        return torch.squeeze(q)
    
    def forward_with_batch(self, state, action):
        
        if self.mlp_shape is not None:
            rnn_input = state[:, :self.rnn_shape[0] * self.rnn_shape[1]]
            mlp_input = state[:, self.rnn_shape[0] * self.rnn_shape[1]:]
        else:
            rnn_input = state
            mlp_input = None
    
        # Reshape rnn_input to (batch_size, time, features)
        rnn_input = rnn_input.view(-1, self.rnn_shape[0], self.rnn_shape[1])

        mlp_input = torch.cat((mlp_input, action), dim=1) # dim 0 is batch dimension
        
        return self.model(rnn_input, mlp_input)
    
    def forward_without_batch(self, state, action):
        if self.mlp_shape is not None:
            rnn_input = state[:self.rnn_shape[0] * self.rnn_shape[1]]
            mlp_input = state[self.rnn_shape[0] * self.rnn_shape[1:]:]
        else:
            rnn_input = state
            mlp_input = None
        
        # Reshape rnn_input to (time, features)
        rnn_input = rnn_input.view(self.rnn_shape[0], self.rnn_shape[1])

        mlp_input = torch.cat((mlp_input, action), dim=0) # no batch dimension  
        
        return self.model(rnn_input, mlp_input)



"""Dict of agent classes and (standard) agent names to allow for dynamic loading of agents."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/40_base_agents/10_AGENT_CLASSES.ipynb.

# %% auto 0
__all__ = ['AGENT_CLASSES']

# %% ../../nbs/40_base_agents/10_AGENT_CLASSES.ipynb 3
AGENT_CLASSES = {
    "RandomAgent": "ddopai.agents.saa.SAA",

    "SAA": "ddopai.agents.newsvendor.saa.NewsvendorSAAagent",
    "wSAA": "ddopai.agents.newsvendor.saa.NewsvendorRFwSAAagent",
    "RFwSAA": "ddopai.agents.newsvendor.saa.NewsvendorRFwSAAagent",

    "XGB": "ddopai.agents.newsvendor.erm.NewsvendorXGBAgent",

    "lERM": "ddopai.agents.newsvendor.erm.NewsvendorlERMAgent",
    "DLNV": "ddopai.agents.newsvendor.erm.NewsvendorDLAgent",
    "DLNVRNN": "ddopai.agents.newsvendor.erm.NewsvendorDLRNNAgent",
    "DLNVTransformer": "ddopai.agents.newsvendor.erm.NewsvendorDLTransformerAgent",

    "lERMMeta": "ddopai.agents.newsvendor.erm.NewsvendorlERMMetaAgent",
    "DLNVMeta": "ddopai.agents.newsvendor.erm.NewsvendorDLMetaAgent",
    "DLNVRNNMeta": "ddopai.agents.newsvendor.erm.NewsvendorDLRNNMetaAgent",
    "DLNVTransformerMeta": "ddopai.agents.newsvendor.erm.NewsvendorDLTransformerMetaAgent",

    "SAC": "ddopai.agents.rl.sac.SACAgent",
    "SACRNN": "ddopai.agents.rl.sac.SACRNNAgent",

    "TD3": "ddopai.agents.rl.td3.TD3Agent",
    "PPO": "ddopai.agents.rl.ppo.PPOAgent",
}



"""Newsvendor agents based on Empirical Risk Minimization (ERM) principles."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb.

# %% auto 0
__all__ = ['NewsvendorXGBAgent', 'SGDBaseAgent', 'NVBaseAgent', 'NewsvendorlERMAgent', 'NewsvendorDLAgent', 'BaseMetaAgent',
           'NewsvendorlERMMetaAgent', 'NewsvendorDLMetaAgent', 'NewsvendorDLTransformerAgent',
           'NewsvendorDLTransformerMetaAgent']

# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 3
import logging

from abc import ABC, abstractmethod
from typing import Union, Optional, List, Tuple, Literal, Callable, Dict
import numpy as np
import os
from tqdm import tqdm
import time
from IPython import get_ipython

import xgboost as xgb

from ...envs.base import BaseEnvironment
from ..base import BaseAgent
from ...utils import MDPInfo, Parameter, DatasetWrapper, DatasetWrapperMeta
from ...torch_utils.loss_functions import TorchQuantileLoss, TorchPinballLoss
from ...obsprocessors import FlattenTimeDimNumpy
from ...dataloaders.base import BaseDataLoader
from ...ml_utils import LRSchedulerPerStep

import torch

from torchinfo import summary

# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 4
class NewsvendorXGBAgent(BaseAgent):

    """
    Agent solving the Newsvendor problem within the ERM framework (i.e., using quantile regression)
    using the XGBoost library.
    """

    def __init__(self,

                    environment_info: MDPInfo,
                    cu: float | np.ndarray, # underage cost
                    co: float | np.ndarray, # overage cost
                    obsprocessors: Optional[List[object]] = None,
                    agent_name: str | None = "XGBAgent",

                    ### XGB params
                    eta: float = 0.3,
                    gamma: float = 0,
                    max_depth: int = 6,
                    min_child_weight: float = 1,
                    max_delta_step: float = 0,
                    subsample: float = 1,
                    sampling_method: str = "uniform",
                    colsample_bytree: float = 1,
                    colsample_bylevel: float = 1,
                    colsample_bynode: float = 1,
                    lambda_: float = 1,
                    alpha: float = 0,
                    tree_method: str = "auto",
                    scale_pos_weight: float = 1,
                    # updater will always use default
                    refresh_leaf: int = 1,
                    # process type will always use default
                    grow_policy: str = "depthwise",
                    max_leaves: int = 0,
                    max_bin: int = 256,
                    num_parallel_tree: int = 1,
                    multi_strategy: str = "one_output_per_tree",
                    max_cached_hist_node: int = 65536,

                    ### General params
                    nthread: int = 1,
                    device: str = "CPU"):

        # if float, convert to array
        cu = self.convert_to_numpy_array(cu)
        co = self.convert_to_numpy_array(co)

        self.sl = cu / (cu + co)
        self.fitted = False

        self.model = xgb.XGBRegressor(
            objective = "reg:quantileerror",
            quantile_alpha = self.sl[0],

            eta=eta,
            gamma=gamma,
            max_depth=max_depth,
            min_child_weight=min_child_weight,
            max_delta_step=max_delta_step,
            subsample=subsample,
            sampling_method=sampling_method,
            colsample_bytree=colsample_bytree,
            colsample_bylevel=colsample_bylevel,
            colsample_bynode=colsample_bynode,
            lambda_=lambda_,
            alpha=alpha,
            tree_method=tree_method,
            scale_pos_weight=scale_pos_weight,
            # updater will always use default
            refresh_leaf=refresh_leaf,
            # process type will always use default
            grow_policy=grow_policy,
            max_leaves=max_leaves,
            max_bin=max_bin,
            num_parallel_tree=num_parallel_tree,
            multi_strategy=multi_strategy,
            max_cached_hist_node=max_cached_hist_node,
            nthread=nthread,
            device=device
        )
        
        super().__init__(environment_info = environment_info, obsprocessors = obsprocessors, agent_name = agent_name)

        # check if FlattenTimeDimNumpy in obsprocessors
        if not any(isinstance(p, FlattenTimeDimNumpy) for p in self.obsprocessors):
            self.add_obsprocessor(FlattenTimeDimNumpy(allow_2d=True))

    def fit(self,
            X: np.ndarray, # features will be ignored
            Y: np.ndarray) -> None:

        """

        Fit the agent to the data by training the XGBoost model.

        """

        if X.ndim == 3:
            X = X.reshape(X.shape[0], -1)
        self.model.fit(X, Y)
        self.fitted = True
    
    def draw_action_(self, 
                    observation: np.ndarray) -> np.ndarray: #
        """

        Draw an action from the model given an observation.
        
        """


        if self.fitted == False:
            return np.array([0.0])

        return self.model.predict(observation)

    
    def save(self,
            path: str,  # The directory where the file will be saved.
            overwrite: bool=True):  # Allow overwriting; if False, a FileExistsError will be raised if the file exists.

        """
        Save the XGBoost model to a file in the specified directory.
        """

        if not hasattr(self, 'fitted') or not self.fitted:
            raise ValueError("Agent has not been fitted yet. Train the model before saving.")
        
        if not hasattr(self, 'model') or self.model is None:
            raise ValueError("Agent has no model to save. The model might not have been trained or initialized.")

        os.makedirs(path, exist_ok=True)

        full_path = os.path.join(path, "xgb_model.json")  # XGBoost models can be saved as .json or .model

        if os.path.exists(full_path):
            if not overwrite:
                raise FileExistsError(f"The file {full_path} already exists and will not be overwritten.")
            else:
                logging.warning(f"Overwriting file {full_path}")

        self.model.save_model(full_path)

    def load(self, path: str):
        full_path = os.path.join(path, "xgb_model.json")
        
        if not os.path.exists(full_path):
            raise FileNotFoundError(f"The file {full_path} does not exist.")
        
        try:
            self.model = xgb.XGBRegressor()  # Use XGBRegressor instead of Booster
            self.model.load_model(full_path)
            self.fitted = True
            logging.info(f"Model loaded successfully from {full_path}")
        except Exception as e:
            raise ValueError(f"An error occurred while loading the model: {e}")



# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 5
class SGDBaseAgent(BaseAgent):

    """
    Base class for Agents that are trained using Stochastic Gradient Descent (SGD) on PyTorch models.
    """

    # TODO: Remove input shapes as input end get from MDPInfo

    train_mode = "epochs_fit"
    
    def __init__(self, 
            environment_info: MDPInfo,
            dataloader: BaseDataLoader,
            input_shape: Tuple,
            output_shape: Tuple,
            dataset_params: Optional[dict] = None, # parameters needed to convert the dataloader to a torch dataset
            dataloader_params: Optional[dict] = None, # default: {"batch_size": 32, "shuffle": True}
            optimizer_params: Optional[dict] = None,  # default: {"optimizer": "Adam", "lr": 0.01, "weight_decay": 0.0}
            learning_rate_scheduler_params: Dict | None = None, # default: None. If dict, then first key is "scheduler" and the rest are the parameters
            obsprocessors: Optional[List] = None,     # default: []
            device: str = "cpu", # "cuda" or "cpu"
            agent_name: str | None = None,
            test_batch_size: int = 1024,
            receive_batch_dim: bool = False,
            ):
        
        # Initialize default values for mutable arguments
        optimizer_params = optimizer_params or {"optimizer": "Adam", "lr": 0.01, "weight_decay": 0.0}
        dataloader_params = dataloader_params or {"batch_size": 32, "shuffle": True}
        dataset_params = dataset_params or {}

        self.device = self.set_device(device)
        
        self.set_dataloader(dataloader, dataset_params, dataloader_params)

        self.set_model(input_shape, output_shape)
        self.loss_function_params=None # default
        self.set_loss_function()
        self.set_optimizer(optimizer_params)
        self.set_learning_rate_scheduler(learning_rate_scheduler_params)
        self.test_batch_size = test_batch_size

        super().__init__(environment_info = environment_info, obsprocessors = obsprocessors, agent_name = agent_name, receive_batch_dim = receive_batch_dim)

        batch_dim = 1
        logging.info("Network architecture:")
        if logging.getLogger().isEnabledFor(logging.INFO):

            self.model.eval()
            if any(isinstance(obsprocessor, FlattenTimeDimNumpy) for obsprocessor in self.obsprocessors):
                input_size = (batch_dim, int(np.prod(input_shape)))
            else:
                input_size = (batch_dim, *input_shape)

            input_tensor = torch.randn(*input_size).to(self.device)
            input_tuple = (input_tensor,)

            if get_ipython() is not None:
                print(summary(self.model, input_data=input_tuple, device=self.device))
            else:
                summary(self.model, input_data=input_tuple, device=self.device)
            time.sleep(0.2)

        self.to(self.device)

    def set_device(self, device: str):

        """ Set the device for the model """

        if device == "cuda":
            if torch.cuda.is_available():
                return "cuda"
            else:
                logging.warning("CUDA is not available. Using CPU instead.")
                return "cpu"
        elif device == "cpu":
            return "cpu"
        else:
            raise ValueError(f"Device {device} not currently not supported, use 'cuda' or 'cpu'")


    def set_dataloader(self,
                        dataloader: BaseDataLoader,
                        dataset_params: dict,
                        dataloader_params: dict, # dict with keys: batch_size, shuffle
                        ) -> None: 

        """
        Set the dataloader for the agent by wrapping it into a Torch Dataset
        
        """

        # check if class already have a dataloader
        if not hasattr(self, 'dataloader'):

            dataset = DatasetWrapper(dataloader, **dataset_params)
            self.dataloader = torch.utils.data.DataLoader(dataset, **dataloader_params)

    @abstractmethod
    def set_loss_function(self):
        """ Set loss function for the model """
        pass

    @abstractmethod
    def set_model(self, input_shape: Tuple, output_shape: Tuple):
        """ Set the model for the agent """
        pass

    def set_optimizer(self, optimizer_params: dict): # dict with keys: optimizer, lr, weight_decay
        
        """ Set the optimizer for the model """

        if not hasattr(self, 'optimizer'):
            
            optimizer = optimizer_params["optimizer"]
            optimizer_params_copy = optimizer_params.copy()
            del optimizer_params_copy["optimizer"]

            if optimizer == "Adam":
                self.optimizer = torch.optim.Adam(self.model.parameters(), **optimizer_params_copy)
            elif optimizer == "SGD":
                self.optimizer = torch.optim.SGD(self.model.parameters(), **optimizer_params_copy)
            elif optimizer == "RMSprop":
                self.optimizer = torch.optim.RMSprop(self.model.parameters(), **optimizer_params_copy)
            else:
                raise ValueError(f"Optimizer {optimizer} not supported")
        
    def set_learning_rate_scheduler(self, learning_rate_scheduler_params): #
        """ Set learning rate scheudler (can be None) """

        if learning_rate_scheduler_params is not None:

            params = learning_rate_scheduler_params.copy()
            scheduler_type = params["scheduler"]
            del params["scheduler"]
            if scheduler_type == "LRSchedulerPerStep":
                self.learning_rate_scheduler = LRSchedulerPerStep(self.optimizer, **params)
            else:
                raise ValueError(f"Learning rate scheduler {scheduler_type} not supported")

        else:
            self.learning_rate_scheduler = None

    def fit_epoch(self):

        """ Fit the model for one epoch using the dataloader """

        device = next(self.model.parameters()).device
        self.model.train()
        total_loss=0

        for i, output in enumerate(tqdm(self.dataloader)):
            
            if len(output)==3:
                X, y, loss_function_params = output
            else:
                X, y = output
                loss_function_params = None

            # convert X and y to float32
            X = X.type(torch.float32)
            y = y.type(torch.float32)
            
            X, y = X.to(device), y.to(device)

            self.optimizer.zero_grad()

            y_pred = self.model(X)

            if loss_function_params is not None:
                loss = self.loss_function(y_pred, y, **loss_function_params)
            elif self.loss_function_params is not None:
                loss = self.loss_function(y_pred, y, **self.loss_function_params)
            else:
                loss = self.loss_function(y_pred, y)

            loss.backward()
            self.optimizer.step()

            if self.learning_rate_scheduler is not None:
                self.learning_rate_scheduler.step()
        
            total_loss += loss.item()
        
        self.model.eval()
        
        return total_loss

    def draw_action_(self, observation: np.ndarray) -> np.ndarray: #
        
        """ 
        Draw an action based on the fitted model (see predict method)
        """
        
        action = self.predict(observation)
        
        return action
    
    @staticmethod
    def split_into_batches(X: np.ndarray, batch_size: int) -> List[np.ndarray]: #
        """ Split the input into batches of the specified size """
        return [X[i:i+batch_size] for i in range(0, len(X), batch_size)]

    def predict(self, X: np.ndarray) -> np.ndarray: #
        """ Do one forward pass of the model and return the prediction """

        device = next(self.model.parameters()).device
        self.model.eval()

        batches = self.split_into_batches(X, self.test_batch_size)

        y_pred_full = []
        for batch in batches:

            X = batch

            X = torch.tensor(X, dtype=torch.float32)
            X = X.to(device)

            with torch.no_grad():

                y_pred = self.model(X)

                # check if y_pred is not finite:
                if not torch.all(torch.isfinite(y_pred)):

                    print(y_pred)

                    # check if X is not finite:
                    if not torch.all(torch.isfinite(X)):

                        print("X is not finite")
                        print("total X_shape: ", X.shape)
                        print("non-finite indices: ", torch.nonzero(~torch.isfinite(X)))
                        print(X)


                    raise ValueError("Predicted values are not finite")

            y_pred = y_pred.cpu().numpy()

            y_pred_full.append(y_pred)
        
        y_pred_full = np.concatenate(y_pred_full, axis=0)

        return y_pred_full

    def train(self):
        """set the internal state of the agent and its model to train"""
        self.mode = "train"
        self.model.train()

    def eval(self):
        """set the internal state of the agent and its model to eval"""
        self.mode = "eval"
        self.model.eval()

    def to(self, device: str): #
        """Move the model to the specified device"""
        self.model.to(device)

    def save(self,
                path: str, # The directory where the file will be saved.
                overwrite: bool=True): # Allow overwriting; if False, a FileExistsError will be raised if the file exists.
        
        """
        Save the PyTorch model to a file in the specified directory.

        """
        
        if not hasattr(self, 'model') or self.model is None:
            raise AttributeError("Model is not defined in the class.")

        # Create the directory path if it does not exist
        os.makedirs(path, exist_ok=True)

        # Construct the file path using os.path.join for better cross-platform compatibility
        full_path = os.path.join(path, "model.pth")

        if os.path.exists(full_path):
            if not overwrite:
                raise FileExistsError(f"The file {full_path} already exists and will not be overwritten.")
            else:
                logging.debug(f"Overwriting file {full_path}") # Only log with info as during training we will continuously overwrite the model
        
        # Save the model's state_dict using torch.save
        torch.save(self.model.state_dict(), full_path)
        logging.debug(f"Model saved successfully to {full_path}")

    def load(self, path: str): # Only the path to the folder is needed, not the file itself
 
        """
        Load the PyTorch model from a file.
        """
        
        if not hasattr(self, 'model') or self.model is None:
            raise AttributeError("Model is not defined in the class.")

        # Construct the file path
        full_path = os.path.join(path, "model.pth")

        if not os.path.exists(full_path):
            raise FileNotFoundError(f"The file {full_path} does not exist.")

        try:
            # Load the model's state_dict using torch.load
            if self.device == "cuda":
                self.model.load_state_dict(torch.load(full_path))
            else:
                self.model.load_state_dict(torch.load(full_path, map_location=torch.device('cpu')))
            logging.debug(f"Model loaded successfully from {full_path}")
        except Exception as e:
            raise RuntimeError(f"An error occurred while loading the model: {e}")
    

# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 21
class NVBaseAgent(SGDBaseAgent):

    """
    Base agent for the Newsvendor problem implementing
    the loss function for the Empirical Risk Minimization (ERM) approach
    based on quantile loss.
    """

    def __init__(self, 
                environment_info: MDPInfo,
                dataloader: BaseDataLoader,
                cu: np.ndarray | Parameter,
                co: np.ndarray | Parameter,
                input_shape: Tuple,
                output_shape: Tuple,
                optimizer_params: dict | None = None,  # default: {"optimizer": "Adam", "lr": 0.01, "weight_decay": 0.0}
                learning_rate_scheduler_params = None,  # TODO: add base class for learning rate scheduler for typing
                dataset_params: dict | None = None, # parameters needed to convert the dataloader to a torch dataset
                dataloader_params: dict | None = None,  # default: {"batch_size": 32, "shuffle": True}
                obsprocessors: list | None = None,      # default: []
                device: str = "cpu", # "cuda" or "cpu"
                agent_name: str | None = None,
                test_batch_size: int = 1024,
                receive_batch_dim: bool = False,
                loss_function: Literal["quantile", "pinball"] = "quantile", 
                ):

        cu = self.convert_to_numpy_array(cu)
        co = self.convert_to_numpy_array(co)
        
        self.sl = cu / (cu + co) # ensure this works if cu and co are Parameters
        self.cu = cu
        self.co = co

        self.loss_function = loss_function


        super().__init__(
            environment_info=environment_info,
            dataloader=dataloader,
            input_shape=input_shape,
            output_shape=output_shape,
            optimizer_params=optimizer_params,
            learning_rate_scheduler_params=learning_rate_scheduler_params,
            dataset_params=dataset_params,
            dataloader_params=dataloader_params,
            obsprocessors=obsprocessors,
            device=device,
            agent_name=agent_name,
            test_batch_size=test_batch_size,
            receive_batch_dim=receive_batch_dim,
        )   
        
    def set_loss_function(self):
        
        """Set the loss function for the model to the quantile loss. For training
        the model uses quantile loss and not the pinball loss with specific cu and 
        co values to ensure similar scale of the feedback signal during training."""

        if self.loss_function == "quantile":
            self.loss_function_params = {"quantile": self.sl}
            self.loss_function = TorchQuantileLoss(reduction="mean")
            logging.debug(f"Loss function set to {self.loss_function}")

        elif self.loss_function == "pinball":
            self.loss_function_params = {"underage": self.cu, "overage": self.co}
            self.loss_function = TorchPinballLoss(reduction="mean")
            logging.debug(f"Loss function set to {self.loss_function}")

        else:
            raise ValueError(f"Loss function {self.loss_function} not supported")

# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 24
class NewsvendorlERMAgent(NVBaseAgent):

    """
    Newsvendor agent implementing Empirical Risk Minimization (ERM) approach 
    based on a linear (regression) model. Note that this implementation finds
    the optimal regression parameters via SGD.

    """

    def __init__(self, 
                environment_info: MDPInfo,
                dataloader: BaseDataLoader,
                cu: np.ndarray | Parameter,
                co: np.ndarray | Parameter,
                input_shape: Tuple,
                output_shape: Tuple,
                optimizer_params: dict | None = None,  # default: {"optimizer": "Adam", "lr": 0.01, "weight_decay": 0.0}
                learning_rate_scheduler_params = None,  # TODO: add base class for learning rate scheduler for typing
                model_params: dict | None = None,  # default: {"relu_output": False}
                dataset_params: dict | None = None, # parameters needed to convert the dataloader to a torch dataset
                dataloader_params: dict | None = None,  # default: {"batch_size": 32, "shuffle": True}
                obsprocessors: list | None = None,  # default: []
                device: str = "cpu",  # "cuda" or "cpu"
                agent_name: str | None = "lERM",
                test_batch_size: int = 1024,
                receive_batch_dim: bool = False,
                loss_function: Literal["quantile", "pinball"] = "quantile", 
                ):

        # Handle mutable defaults unique to this class
        default_model_params = {
            "relu_output": False
            }

        self.model_params = self.update_model_params(default_model_params, model_params or {})

        super().__init__(
            environment_info=environment_info,
            dataloader=dataloader,
            cu=cu,
            co=co,
            input_shape=input_shape,
            output_shape=output_shape,
            optimizer_params=optimizer_params,
            learning_rate_scheduler_params=learning_rate_scheduler_params,
            dataloader_params=dataloader_params,
            dataset_params=dataset_params,
            obsprocessors=obsprocessors,
            device=device,
            agent_name=agent_name,
            test_batch_size=test_batch_size,
            receive_batch_dim=receive_batch_dim,
            loss_function=loss_function,
        )
    def set_model(self, input_shape, output_shape):

        """Set the model for the agent to a linear model"""

        from ddopai.approximators import LinearModel

        # flatten time dim of input
        print("input shape", input_shape)
        input_size = np.prod(input_shape)
        output_size = output_shape[0]

        self.model = LinearModel(input_size=input_size, output_size=output_size, **self.model_params)

# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 31
class NewsvendorDLAgent(NVBaseAgent):

    """
    Newsvendor agent implementing Empirical Risk Minimization (ERM) approach 
    based on a deep learning model. 
    """

    def __init__(self, 
                environment_info: MDPInfo,
                dataloader: BaseDataLoader,
                cu: np.ndarray | Parameter,
                co: np.ndarray | Parameter,
                input_shape: Tuple,
                output_shape: Tuple,
                learning_rate_scheduler_params: Dict | None = None,  
                
                # parameters in yaml file
                optimizer_params: dict | None = None,  # default: {"optimizer": "Adam", "lr": 0.01, "weight_decay": 0.0}
                model_params: dict | None = None,  # default: {"hidden_layers": [64, 64], "drop_prob": 0.0, "batch_norm": False, "relu_output": False}
                dataloader_params: dict | None = None,  # default: {"batch_size": 32, "shuffle": True}
                dataset_params: dict | None = None, # parameters needed to convert the dataloader to a torch dataset
                device: str = "cpu", # "cuda" or "cpu"

                obsprocessors: list | None = None,  # default: []
                agent_name: str | None = "DLNV",
                test_batch_size: int = 1024,
                receive_batch_dim: bool = False,
                loss_function: Literal["quantile", "pinball"] = "quantile",
                ):

        # Handle mutable defaults unique to this class
        default_model_params = {
            "hidden_layers": [64, 64],
            "drop_prob": 0.0,
            "batch_norm": False,
            "relu_output": False
            }

        self.model_params = self.update_model_params(default_model_params, model_params or {})

        super().__init__(
            environment_info=environment_info,
            dataloader=dataloader,
            cu=cu,
            co=co,
            input_shape=input_shape,
            output_shape=output_shape,
            optimizer_params=optimizer_params,
            learning_rate_scheduler_params=learning_rate_scheduler_params,
            dataloader_params=dataloader_params,
            dataset_params=dataset_params,
            obsprocessors=obsprocessors,
            device=device,
            agent_name=agent_name,
            test_batch_size=test_batch_size,
            receive_batch_dim=receive_batch_dim,
            loss_function=loss_function,
        )
        
    def set_model(self, input_shape, output_shape):
        
        """Set the model for the agent to an MLP"""

        # flatten time dim of input
        print("input shape", input_shape)
        input_size = np.prod(input_shape)
        output_size = output_shape[0]

        from ddopai.approximators import MLP
        self.model = MLP(input_size=input_size, output_size=output_size, **self.model_params)

# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 37
class BaseMetaAgent():

    def set_meta_dataloader(
        self, 
        dataloader: BaseDataLoader,
        dataset_params: dict, # parameters needed to convert the dataloader to a torch dataset
        dataloader_params: dict, # dict with keys: batch_size, shuffle
        ) -> None:

        """ """

        dataset = DatasetWrapperMeta(dataloader, **dataset_params)

        self.dataloader = torch.utils.data.DataLoader(dataset, **dataloader_params)

# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 38
class NewsvendorlERMMetaAgent(NewsvendorlERMAgent, BaseMetaAgent):

    """
    Newsvendor agent implementing Empirical Risk Minimization (ERM) approach 
    based on a linear (regression) model. In addition to the features, the agent
    also gets the sl as input to be able to forecast the optimal order quantity
    for different sl values. Depending on the training pipeline, this model can be 
    adapted to become a full meta-learning algorithm cross products and cross sls.

    """

    def __init__(self, 

                # Parameters for lERM agent
                environment_info: MDPInfo,
                dataloader: BaseDataLoader,
                cu: np.ndarray | Parameter,
                co: np.ndarray | Parameter,
                input_shape: Tuple,
                output_shape: Tuple,
                optimizer_params: dict | None = None,  # default: {"optimizer": "Adam", "lr": 0.01, "weight_decay": 0.0}
                learning_rate_scheduler_params = None,  # TODO: add base class for learning rate scheduler for typing
                model_params: dict | None = None,  # default: {"relu_output": False}
                dataset_params: dict | None = None, # parameters needed to convert the dataloader to a torch dataset
                dataloader_params: dict | None = None,  # default: {"batch_size": 32, "shuffle": True}
                obsprocessors: list | None = None,  # default: []
                device: str = "cpu",  # "cuda" or "cpu"
                agent_name: str | None = "lERMMeta",
                test_batch_size: int = 1024,
                receive_batch_dim: bool = False,
                loss_function: Literal["quantile", "pinball"] = "quantile",
                ):

        self.set_meta_dataloader(dataloader, dataset_params, dataloader_params)

        super().__init__(
            environment_info=environment_info,
            dataloader=dataloader,
            cu=cu,
            co=co,
            input_shape=input_shape,
            output_shape=output_shape,
            optimizer_params=optimizer_params,
            learning_rate_scheduler_params=learning_rate_scheduler_params,
            model_params=model_params,
            dataloader_params=dataloader_params,
            obsprocessors=obsprocessors,
            device=device,
            agent_name=agent_name,
            test_batch_size=test_batch_size,
            receive_batch_dim = receive_batch_dim,
            loss_function=loss_function,
        )

# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 39
class NewsvendorDLMetaAgent(NewsvendorDLAgent, BaseMetaAgent):

    """
    Newsvendor agent implementing Empirical Risk Minimization (ERM) approach 
    based on a Neural Network. In addition to the features, the agent
    also gets the sl as input to be able to forecast the optimal order quantity
    for different sl values. Depending on the training pipeline, this model can be 
    adapted to become a full meta-learning algorithm cross products and cross sls.

    """

    def __init__(self, 
                environment_info: MDPInfo,
                dataloader: BaseDataLoader,
                cu: np.ndarray | Parameter,
                co: np.ndarray | Parameter,
                input_shape: Tuple,
                output_shape: Tuple,
                learning_rate_scheduler_params = None,  # TODO: add base class for learning rate scheduler for typing
                
                # parameters in yaml file
                optimizer_params: dict | None = None,  # default: {"optimizer": "Adam", "lr": 0.01, "weight_decay": 0.0}
                model_params: dict | None = None,  # default: {"hidden_layers": [64, 64], "drop_prob": 0.0, "batch_norm": False, "relu_output": False}
                dataset_params: dict | None = None, # parameters needed to convert the dataloader to a torch dataset
                dataloader_params: dict | None = None,  # default: {"batch_size": 32, "shuffle": True}
                device: str = "cpu", # "cuda" or "cpu"

                obsprocessors: list | None = None,  # default: []
                agent_name: str | None = "DLNV",
                test_batch_size: int = 1024,
                receive_batch_dim: bool = False,
                loss_function: Literal["quantile", "pinball"] = "quantile",
                ):

        self.set_meta_dataloader(dataloader, dataset_params, dataloader_params)

        super().__init__(
            environment_info=environment_info,
            dataloader=dataloader,
            cu=cu,
            co=co,
            input_shape=input_shape,
            output_shape=output_shape,
            learning_rate_scheduler_params=learning_rate_scheduler_params,

            optimizer_params=optimizer_params,
            model_params=model_params,
            dataloader_params=dataloader_params,
            device=device,

            obsprocessors=obsprocessors,
            agent_name=agent_name,
            test_batch_size=test_batch_size,
            receive_batch_dim=receive_batch_dim,
            loss_function=loss_function,
        )


# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 40
class NewsvendorDLTransformerAgent(NVBaseAgent):

    """
    Newsvendor agent implementing Empirical Risk Minimization (ERM) approach 
    based on a deep learning model with a Transformer architecture.
    """

    def __init__(self, 
                environment_info: MDPInfo,
                dataloader: BaseDataLoader,
                cu: np.ndarray | Parameter,
                co: np.ndarray | Parameter,
                input_shape: Tuple,
                output_shape: Tuple,
                learning_rate_scheduler_params: Dict | None = None,
                
                # parameters in yaml file
                optimizer_params: dict | None = None,  # default: {"optimizer": "Adam", "lr": 0.01, "weight_decay": 0.0}
                model_params: dict | None = None,  # default: {"max_context_length": 128, "n_layer": 3, "n_head": 8, "n_embd_per_head": 32, "rope_scaling": None, "min_multiple": 256, "gating": True, "drop_prob": 0.0, "final_activation": "identity"}
                dataset_params: dict | None = None, # parameters needed to convert the dataloader to a torch dataset
                dataloader_params: dict | None = None,  # default: {"batch_size": 32, "shuffle": True}
                device: str = "cpu", # "cuda" or "cpu"

                obsprocessors: list | None = None,  # default: []
                agent_name: str | None = "DLNV",
                test_batch_size: int = 1024,
                receive_batch_dim: bool = False,
                loss_function: Literal["quantile", "pinball"] = "quantile",
                ):

        # Handle mutable defaults unique to this class
        default_model_params = {
            "max_context_length": 128,
            "n_layer": 3,
            "n_head": 8,
            "n_embd_per_head": 32,
            "rope_scaling": None,

            "min_multiple": 256,
            "gating": True,

            "drop_prob": 0.0,
            "final_activation": "identity",
            }
        
        self.model_params = self.update_model_params(default_model_params, model_params or {})


        super().__init__(
            environment_info=environment_info,
            dataloader=dataloader,
            cu=cu,
            co=co,
            input_shape=input_shape,
            output_shape=output_shape,
            optimizer_params=optimizer_params,
            learning_rate_scheduler_params=learning_rate_scheduler_params,
            dataset_params=dataset_params,
            dataloader_params=dataloader_params,
            obsprocessors=obsprocessors,
            device=device,
            agent_name=agent_name,
            test_batch_size=test_batch_size,
            receive_batch_dim=receive_batch_dim,
            loss_function=loss_function,
        )
         
    def set_model(self, input_shape, output_shape):
        
        """Set the model for the agent to an MLP"""

        if len(input_shape) == 1:
            raise ValueError("Input shape must be at least 2D for Transformer model")

        output_size = output_shape[0]

        from ddopai.approximators import Transformer
        self.model = Transformer(input_size=input_shape, output_size=output_size, **self.model_params)

# %% ../../../nbs/41_NV_agents/11_NV_erm_agents.ipynb 41
class NewsvendorDLTransformerMetaAgent(NewsvendorDLTransformerAgent, BaseMetaAgent):

    """
    Newsvendor agent implementing Empirical Risk Minimization (ERM) approach 
    based on a Neural Network using the attention mechanism. In addition to the features,
    the agent also gets the sl as input to be able to forecast the optimal order quantity
    for different sl values. Depending on the training pipeline, this model can be 
    adapted to become a full meta-learning algorithm cross products and cross sls.

    """

    def __init__(self, 
                environment_info: MDPInfo,
                dataloader: BaseDataLoader,
                cu: np.ndarray | Parameter,
                co: np.ndarray | Parameter,
                input_shape: Tuple,
                output_shape: Tuple,
                learning_rate_scheduler_params: Dict | None = None, 
                
                # parameters in yaml file
                optimizer_params: dict | None = None,  # default: {"optimizer": "Adam", "lr": 0.01, "weight_decay": 0.0}
                model_params: dict | None = None,  # default: {"hidden_layers": [64, 64], "drop_prob": 0.0, "batch_norm": False, "relu_output": False}
                dataset_params: dict | None = None, # parameters needed to convert the dataloader to a torch dataset
                dataloader_params: dict | None = None,  # default: {"batch_size": 32, "shuffle": True}
                device: str = "cpu", # "cuda" or "cpu"

                obsprocessors: list | None = None,  # default: []
                agent_name: str | None = "DLNV",
                test_batch_size: int = 1024,
                receive_batch_dim: bool = False,
                loss_function: Literal["quantile", "pinball"] = "quantile",
                ):

        self.set_meta_dataloader(dataloader, dataset_params, dataloader_params)

        super().__init__(
            environment_info=environment_info,
            dataloader=dataloader,
            cu=cu,
            co=co,
            input_shape=input_shape,
            output_shape=output_shape,
            learning_rate_scheduler_params=learning_rate_scheduler_params,

            optimizer_params=optimizer_params,
            model_params=model_params,
            dataloader_params=dataloader_params,
            device=device,

            obsprocessors=obsprocessors,
            agent_name=agent_name,
            test_batch_size=test_batch_size,
            receive_batch_dim=receive_batch_dim,
            loss_function=loss_function,
        )




"""Base environment with some basic funcitons"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/21_envs_inventory/00_inventory_utils.ipynb.

# %% auto 0
__all__ = ['OrderPipeline']

# %% ../../../nbs/21_envs_inventory/00_inventory_utils.ipynb 3
import logging


from abc import ABC, abstractmethod
from typing import Union, Tuple, List, Literal

from ..base import BaseEnvironment
from ...utils import Parameter, MDPInfo
from ...dataloaders.base import BaseDataLoader
from ...loss_functions import pinball_loss
from ...utils import set_param, Parameter


import gymnasium as gym

import numpy as np
import time

# %% ../../../nbs/21_envs_inventory/00_inventory_utils.ipynb 4
class OrderPipeline():
   
    """
    Class to handle the order pipeline in the inventory environments. It is used to keep track of the orders
    that are placed. It can account for fixed and variable lead times.
    
    """

    def __init__(self, 

        num_units: int,  # number of units (SKUs)
        lead_time_mean: Parameter | np.ndarray | List | int | float,  # mean lead time
        lead_time_stochasticity: Literal["fixed", "gamma", "normal_absolute", "normal_relative"] = "fixed", # "fixed", "gamma", "normal_absolute", "normal_relative"
        lead_time_variance: Parameter | np.ndarray | List | int | float | None = None,  # variance of the lead time
        max_lead_time: list[object] | None = None,  # maximum lead time in case of stochastic lead times
        min_lead_time: list[object] | None = 1,  # minimum lead time in case of stochastic lead times

        ) -> None:

        self.set_param('num_units', num_units, shape=(1,), new=True)
        self.set_param('lead_time_mean', lead_time_mean, shape=(self.num_units[0],), new=True)
        self.set_param('lead_time_variance', lead_time_variance, shape=(self.num_units[0],), new=True)
        self.lead_time_stochasticity = lead_time_stochasticity
        self.check_stochasticity(max_lead_time)

        if max_lead_time is None:
            self.set_param('max_lead_time', lead_time_mean, shape=(self.num_units[0],), new=True)
        else:
            self.set_param('max_lead_time', max_lead_time, shape=(self.num_units[0],), new=True)
        self.set_param('min_lead_time', min_lead_time, shape=(self.num_units[0],), new=True)

        if self.max_lead_time is None:
            self.max_lead_time = self.lead_time_mean
        if self.min_lead_time is None:
            self.min_lead_time = 1

        self.check_max_min_mean_lt()
  
        self.pipeline = np.zeros((np.max(self.max_lead_time), num_units))
        self.lead_time_realized = np.zeros((np.max(self.max_lead_time), num_units))

    def get_pipeline(self) -> np.ndarray:
        """ Get the current pipeline """

        return self.pipeline

    def reset(self) -> None:
        """ Reset the pipeline """

        self.pipeline = np.zeros((np.max(self.max_lead_time), self.num_units[0]))
        self.lead_time_realized = np.zeros((np.max(self.max_lead_time), self.num_units[0]))


    def step(self, 
        orders: np.ndarray,
        ) -> np.ndarray:
        
        """ Add orders to the pipeline and return the orders that are arriving """


        # print("beginnig pipeline:")
        # print(self.pipeline)
        # print("beginnig lead_time_realized:")
        # print(self.lead_time_realized)


        orders_arriving = self.get_orders_arriving().copy()
        lead_times = self.draw_lead_times().copy()

        # print("orders_arriving:")
        # print(orders_arriving)
        # print("new orders:")
        # print(orders)
        # print("new lead_times:")
        # print(lead_times)


        self.pipeline = np.roll(self.pipeline, -1, axis=0)
        self.lead_time_realized = np.roll(self.lead_time_realized, -1, axis=0)
        self.pipeline[-1, :] = 0
        self.lead_time_realized[-1, :] = 0
        
        self.pipeline[-1, :] = orders.copy()
        self.lead_time_realized[-1, :] = lead_times
        self.lead_time_realized -= 1
        self.lead_time_realized = np.clip(self.lead_time_realized, 0, None)

        # print("ending pipeline:")
        # print(self.pipeline)
        # print("ending lead_time_realized:")
        # print(self.lead_time_realized)

        return orders_arriving

    def get_orders_arriving(self) -> np.ndarray:

        """ Get the orders that are arriving in the current period """

        orders_arriving = np.zeros(self.num_units[0])
        for i in range(self.num_units[0]):
            # check along the pipeline where the lead time is 0

            arriving_indices = np.where(self.lead_time_realized[:, i] == 0)[0]

            orders_arriving[i] = np.sum(self.pipeline[arriving_indices, i])

            self.pipeline[arriving_indices, i] = 0

        return orders_arriving

    def draw_lead_times(self) -> np.ndarray:
        """ Draw lead times for the orders """

        if self.lead_time_stochasticity == "fixed":
            lead_times = self.lead_time_mean
        elif self.lead_time_stochasticity == "gamma":
            lead_times = np.random.gamma(self.lead_time_mean, 1, self.num_units[0])
        elif self.lead_time_stochasticity == "normal_absolute":
            lead_times = np.random.normal(self.lead_time_mean, self.lead_time_variance, self.num_units[0])
        elif self.lead_time_stochasticity == "normal_relative":
            lead_times = np.random.normal(self.lead_time_mean, self.lead_time_mean * self.lead_time_variance, self.num_units[0])
        else:
            raise ValueError("Invalid lead time stochasticity")

        lead_times = np.clip(lead_times, self.min_lead_time, self.max_lead_time)
        lead_times = np.round(lead_times).astype(int)

        return lead_times

    def check_stochasticity(self, max_lead_time):
        """ Check that params for stochastic lead times are set correctly """

        # lead time mean to be set in any case (it will be the determinstic lead time if lead_time_stochasticity is fixed)
        if self.lead_time_mean is None:
            raise ValueError("Lead time mean is not set")
        if self.lead_time_stochasticity == "fixed" or self.lead_time_stochasticity == "gamma":
            if self.lead_time_variance is not None:
                raise ValueError("Lead time variance must be None for fixed lead times (no variance) or gamma lead times (variance is set by the gamma distribution)")
        elif self.lead_time_stochasticity == "normal_absolute" or self.lead_time_stochasticity == "normal_relative":
            if self.lead_time_variance is None:
                raise ValueError("Lead time variance must be set for normal lead times")
        else:
            raise ValueError("Invalid lead time stochasticity")
        
        if self.lead_time_stochasticity != "fixed":
            if max_lead_time is None:
                raise ValueError("Max lead time must be set for stochastic lead times")

    def check_max_min_mean_lt(self):
        if np.any(self.max_lead_time < self.lead_time_mean):
            raise ValueError("Max lead time must be greater than or equal to the lead time mean")
        if np.any(self.min_lead_time < 1):
            raise ValueError("Min lead time must be at least 1")
        if np.any(self.min_lead_time > self.lead_time_mean):
            raise ValueError("Min lead time must be less than or equal to the lead time mean")
        if np.any(self.max_lead_time < self.min_lead_time):
            raise ValueError("Max lead time must be greater than or equal to the min lead time")
        if np.any(self.max_lead_time < 1):
            raise ValueError("Max lead time must be at least 1")
        if np.any(self.lead_time_mean < 1):
            raise ValueError("Lead time mean must be at least 1")
            
        
    def set_param(self,
                        name: str, # name of the parameter (will become the attribute name)
                        input: Parameter | int | float | np.ndarray | List, # input value of the parameter
                        shape: tuple = (1,), # shape of the parameter
                        new: bool = False # whether to create a new parameter or update an existing one
                        ) -> None: #
        
        """
        Set a parameter for the environment. It converts scalar values to numpy arrays and ensures that
        environment parameters are either of the Parameter class of Numpy arrays. If new is set to True, 
        the function will create a new parameter or update an existing one otherwise. If new is set to
        False, the function will raise an error if the parameter does not exist.
        """

        set_param(self, name, input, shape, new)

    @property
    def shape(self) -> Tuple:
        """ Get the shape of the pipeline """

        return self.pipeline.shape




"""Static inventory environment where a decision only affects the next period (Newsvendor problem)"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/21_envs_inventory/20_single_period_envs.ipynb.

# %% auto 0
__all__ = ['NewsvendorEnv', 'NewsvendorEnvVariableSL']

# %% ../../../nbs/21_envs_inventory/20_single_period_envs.ipynb 3
from abc import ABC, abstractmethod
from typing import Union, Tuple, Literal

from ...utils import Parameter, MDPInfo
from ...dataloaders.base import BaseDataLoader
from ...loss_functions import pinball_loss, quantile_loss
from .base import BaseInventoryEnv

import gymnasium as gym

import numpy as np
import time

# %% ../../../nbs/21_envs_inventory/20_single_period_envs.ipynb 4
class NewsvendorEnv(BaseInventoryEnv, ABC):
    
    """
    Class implementing the Newsvendor problem, working for the single- and multi-item case. If underage_cost and overage_cost
    are scalars and there are multiple SKUs, then the same cost is used for all SKUs. If underage_cost and overage_cost are arrays,
    then they must have the same length as the number of SKUs. Num_SKUs can be set as parameter or inferred from the DataLoader.
    """

    def __init__(self,
        underage_cost: Union[np.ndarray, Parameter, int, float] = 1, # underage cost per unit
        overage_cost: Union[np.ndarray, Parameter, int, float] = 1, # overage cost per unit
        q_bound_low: Union[np.ndarray, Parameter, int, float] = 0, # lower bound of the order quantity
        q_bound_high: Union[np.ndarray, Parameter, int, float] = np.inf, # upper bound of the order quantity
        dataloader: BaseDataLoader = None, # dataloader
        num_SKUs: Union[int] = None, # if None it will be inferred from the DataLoader
        gamma: float = 1, # discount factor
        horizon_train: int | str = "use_all_data", # if "use_all_data" then horizon is inferred from the DataLoader
        postprocessors: list[object] | None = None,  # default is empty list
        mode: str = "train", # Initial mode (train, val, test) of the environment
        return_truncation: str = True # whether to return a truncated condition in step function
    ) -> None:

        self.print=False

        num_SKUs = dataloader.num_units if num_SKUs is None else num_SKUs

        if not isinstance(num_SKUs, int):
            raise ValueError("num_SKUs must be an integer.")
        
        self.set_param("num_SKUs", num_SKUs, shape=(1,), new=True)

        self.set_param("q_bound_low", q_bound_low, shape=(num_SKUs,), new=True)
        self.set_param("q_bound_high", q_bound_high, shape=(num_SKUs,), new=True)

        self.set_observation_space(dataloader.X_shape)
        self.set_action_space(dataloader.Y_shape, low = self.q_bound_low, high = self.q_bound_high)
        
        mdp_info = MDPInfo(self.observation_space, self.action_space, gamma=gamma, horizon=horizon_train)
        
        super().__init__(mdp_info=mdp_info,
                            postprocessors = postprocessors, 
                            mode=mode, return_truncation=return_truncation,
                            underage_cost=underage_cost,
                            overage_cost=overage_cost, 
                            dataloader=dataloader,
                            horizon_train = horizon_train)
        
        

    def step_(self, 
            action: np.ndarray # order quantity
            ) -> Tuple[np.ndarray, float, bool, bool, dict]:

        """
        Step function implementing the Newsvendor logic. Note that the dataloader will return an observation and a demand,
        which will be relevant in the next period. The observation will be returned directly, while the demand will be 
        temporarily stored under self.demand and used in the next step.

        """

        # Most agent give by default a batch dimension which is not needed for a single period action.
        # If action shape size is 2 and the first dimensiion is 1, then remove it
        if action.ndim == 2 and action.shape[0] == 1:
            action = np.squeeze(action, axis=0)  # Remove the first dimension

        cost_per_SKU = self.determine_cost(action)
        reward = -np.sum(cost_per_SKU) # negative because we want to minimize the cost

        terminated = False # in this problem there is no termination condition
        
        info = dict(
            demand=self.demand.copy(),
            action=action.copy(),
            cost_per_SKU=cost_per_SKU.copy()
        )

        # Set index will set the index and return True if the index is out of bounds
        truncated = self.set_index()

        if truncated:

            # observation = np.zeros_like(self.observation_space.sample()) if self.observation_space is not None else None
            # demand = np.zeros_like(self.action_space.sample())

            if self.mode == "test" or self.mode == "val":
                observation, self.demand = None, None
            else:
                observation, self.demand = self.get_observation()

            return observation, reward, terminated, truncated, info
        
        else:

            observation, self.demand = self.get_observation()

            if self.print:
                print("next_period:", self.index+1)
                print("next observation:", observation)
                print("next demand:", self.demand)
                time.sleep(3)

            return observation, reward, terminated, truncated, info

    def determine_cost(self, action: np.ndarray) -> np.ndarray:
        """
        Determine the cost per SKU given the action taken. The cost is the sum of underage and overage costs.
        """
        # Compute the cost per SKU
        return pinball_loss(self.demand, action, self.underage_cost, self.overage_cost)

    def update_cu_co(self, cu=None, co=None):
        # Check if the underage_cost and overage_cost are already set
        if not hasattr(self, "underage_cost") or not hasattr(self, "overage_cost"):
            logging.warning("Underage and overage costs were not set previously, setting them as new parameters.")
            self.set_param("underage_cost", cu, shape=(self.num_SKUs[0],), new=True)
            self.set_param("overage_cost", co, shape=(self.num_SKUs[0],), new=True)
        else:
            # If only cu is provided
            if cu is not None and co is None:
                if len(set(self.overage_cost)) == 1:  # Check if overage_cost is consistent across SKUs
                    self.set_param("overage_cost", self.overage_cost[0], shape=(self.num_SKUs[0],))
                else:
                    raise ValueError("Cannot update cu without updating co when co has heterogeneous values across SKUs.")
                self.set_param("underage_cost", cu, shape=(self.num_SKUs[0],))
            
            # If only co is provided
            elif co is not None and cu is None:
                if len(set(self.underage_cost)) == 1:  # Check if underage_cost is consistent across SKUs
                    self.set_param("underage_cost", self.underage_cost[0], shape=(self.num_SKUs[0],))
                else:
                    raise ValueError("Cannot update co without updating cu when cu has heterogeneous values across SKUs.")
                self.set_param("overage_cost", co, shape=(self.num_SKUs[0],))
            
            # If both cu and co are provided
            elif cu is not None and co is not None:
                self.set_param("underage_cost", cu, shape=(self.num_SKUs[0],))
                self.set_param("overage_cost", co, shape=(self.num_SKUs[0],))
            
        # Update the service level if applicable
        if hasattr(self, "sl"):
            sl = self.underage_cost / (self.underage_cost + self.overage_cost)
            self.set_param("sl", sl, shape=(self.num_SKUs[0],))


# %% ../../../nbs/21_envs_inventory/20_single_period_envs.ipynb 14
class NewsvendorEnvVariableSL(NewsvendorEnv, ABC):
    def __init__(self,

        # Additional parameters:
        sl_bound_low: Union[np.ndarray, Parameter, int, float] = 0.1, # lower bound of the service level during training
        sl_bound_high: Union[np.ndarray, Parameter, int, float] = 0.9, # upper bound of the service level during training
        sl_distribution: Literal["fixed", "uniform"] = "fixed", # distribution of the random service level during training, if fixed then the service level is fixed to sl_test_val
        evaluation_metric: Literal["pinball_loss", "quantile_loss"] = "quantile_loss", # quantile loss is the generic quantile loss (independent of cost levels) while pinball loss uses the specific under- and overage costs
        sl_test_val: Union[np.ndarray, Parameter, int, float] = None, # service level during test and validation, alternatively use cu and co

        underage_cost: Union[np.ndarray, Parameter, int, float] = 1, # underage cost per unit
        overage_cost: Union[np.ndarray, Parameter, int, float] = 1, # overage cost per unit
        q_bound_low: Union[np.ndarray, Parameter, int, float] = 0, # lower bound of the order quantity
        q_bound_high: Union[np.ndarray, Parameter, int, float] = np.inf, # upper bound of the order quantity
        dataloader: BaseDataLoader = None, # dataloader
        num_SKUs: Union[int] = None, # if None it will be inferred from the DataLoader
        gamma: float = 1, # discount factor
        horizon_train: int | str = "use_all_data", # if "use_all_data" then horizon is inferred from the DataLoader
        postprocessors: list[object] | None = None,  # default is empty list
        mode: str = "train", # Initial mode (train, val, test) of the environment
        return_truncation: str = True, # whether to return a truncated condition in step function
        SKUs_in_batch_dimension: bool = True # whether SKUs in the observation space are in the batch dimension (used for meta-learning)
    
    ) -> None:

        # Determine the number of SKUs
        num_SKUs = dataloader.num_units if num_SKUs is None else num_SKUs

        self.set_param("sl_bound_low", sl_bound_low, shape=(1,), new=True)
        self.set_param("sl_bound_high", sl_bound_high, shape=(1,), new=True)
        self.evaluation_metric = evaluation_metric
        self.check_evaluation_metric
        self.sl_distribution = sl_distribution
        self.check_sl_distribution
        self.SKUs_in_batch_dimension = SKUs_in_batch_dimension

        super().__init__(underage_cost=underage_cost,
                        overage_cost=overage_cost,
                        q_bound_low=q_bound_low,
                        q_bound_high=q_bound_high,
                        dataloader=dataloader,
                        num_SKUs=num_SKUs,
                        gamma=gamma,
                        horizon_train=horizon_train,
                        postprocessors=postprocessors,
                        mode=mode,
                        return_truncation=return_truncation)

        if sl_test_val is not None:
            if self.underage_cost is None and self.overage_cost is None:
                self.set_param("sl", sl_test_val, shape=(num_SKUs[0],), new=True)
            else:
                raise ValueError("sl_test_val can only be used when underage_cost and overage_cost are None.")
        else:
            if self.underage_cost is None or self.overage_cost is None:
                raise ValueError("Either sl_test_val or underage_cost and overage_cost must be provided.")
            sl = self.underage_cost / (self.underage_cost + self.overage_cost)
            self.set_param("sl", sl, shape=(self.num_SKUs[0],), new=True)

    def determine_cost(self, action: np.ndarray) -> np.ndarray:
        """
        Determine the cost per SKU given the action taken. The cost is the sum of underage and overage costs.
        """

        # Compute the cost per SKU
        if self.mode == "train": # during training only the service level is relevant
            return quantile_loss(self.demand, action, self.sl_period)
        else:
            if self.evaluation_metric == "pinball_loss":
                return pinball_loss(self.demand, action, self.underage_cost, self.overage_cost)
            elif self.evaluation_metric == "quantile_loss":
                return quantile_loss(self.demand, action, self.sl)

    def set_observation_space(self,
                            shape: tuple, # shape of the dataloader features
                            low: Union[np.ndarray, float] = -np.inf, # lower bound of the observation space
                            high: Union[np.ndarray, float] = np.inf, # upper bound of the observation space
                            samples_dim_included = True # whether the first dimension of the shape input is the number of samples
                            ) -> None:
        
        '''
        Set the observation space of the environment.
        This is a standard function for simple observation spaces. For more complex observation spaces,
        this function should be overwritten. Note that it is assumped that the first dimension
        is n_samples that is not relevant for the observation space.

        '''

        # To handle cases when no external information is available (e.g., parametric NV)
        
        if shape is None:
            self.observation_space = None

        spaces = {}
        if isinstance(shape, tuple):
            if samples_dim_included:
                shape = shape[1:] # assumed that the first dimension is the number of samples
            if self.SKUs_in_batch_dimension:
                shape = (self.num_SKUs[0],) + shape
            spaces["features"] = gym.spaces.Box(low=low, high=high, shape=shape, dtype=np.float32)

        elif feature_shape is None:
            pass

        else:
            raise ValueError("Shape for features must be a tuple or None")

        # TODO check if this is a good desig decision
        if self.SKUs_in_batch_dimension:
            spaces["service_level"] = gym.spaces.Box(low=0, high=1, shape=(1,), dtype=np.float32)
        else:
            spaces["service_level"] = gym.spaces.Box(low=0, high=1, shape=(self.num_SKUs[0],), dtype=np.float32)

        self.observation_space = gym.spaces.Dict(spaces)

    @staticmethod # staticmethod such that the dataloader can also use the funciton
    def draw_parameter(distribution, sl_bound_low, sl_bound_high, samples):
        
        if distribution == "fixed":
            sl = np.random.uniform(sl_bound_low, sl_bound_high, size=(samples,))
        elif distribution == "uniform":
            sl = np.random.uniform(sl_bound_low, sl_bound_high, size=(samples,))
        else:
            raise ValueError("sl_distribution not recognized.")
        
        return sl

    def get_observation(self):
        
        """
        Return the current observation. This function is for the simple case where the observation
        is only an x,y pair. For more complex observations, this function should be overwritten.
        """

        # print("env mode:", self.mode)
        # print("dataloader mode:", self.dataloader.dataset_type)

        X_item, Y_item = self.dataloader[self.index]

        # check if any value in X_item is nan.

        if np.isnan(X_item).any():
            print("X_item is nan")
            # print index
            print("total index:", self.index)
            print("first nan encountered:", np.argwhere(np.isnan(X_item))[0])
        
        if np.isnan(Y_item).any():
            print("Y_item is nan")
            # print index
            print("total index:", self.index)
            print("first nan encountered:", np.argwhere(np.isnan(Y_item))[0])

        # check if any value in X_item is inf.

        if np.isinf(X_item).any():
            print("X_item is inf")
            # print index
            print("total index:", self.index)
            print("first inf encountered:", np.argwhere(np.isinf(X_item))[0])
            
            assert False

        if self.mode == "train":
            sl = self.draw_parameter(self.sl_distribution, self.sl_bound_low, self.sl_bound_high, samples = self.num_SKUs[0])
        else:
            sl = self.sl.copy() # evaluate on fixed sls

        if self.mode != "train":
            if hasattr(self.dataloader, "meta_learn_units") and self.dataloader.meta_learn_units: # dataloaders that train SKU in the batch dimension will put SKU dimension last for validation and test set
                X_item = np.moveaxis(X_item, -1, 0)
 
        self.sl_period = sl # store the service level to assess the action
        
        # print("shape in get observation:", X_item.shape)
        # print("demand in get observation:", Y_item.shape)
        # print("sl in get observation:", sl.shape)

        return {"features": X_item, "service_level": sl}, Y_item

    def check_evaluation_metric(self):
        if self.evaluation_metric not in ["pinball_loss", "quantile_loss"]:
            raise ValueError("evaluation_metric must be either 'pinball_loss' or 'quantile_loss'.")
        if self.evaluation_metric == "pinball_loss" and (self.underage_cost is None or self.overage_cost is None):
            raise ValueError("Underage and overage costs must be provided for pinball loss.")
        if self.evaluation_metric == "quantile_loss" and (self.sl_test_val is None):
            raise ValueError("sl_test_val must be provided for quantile loss.")
    
    def check_sl_distribution(self):
        if self.sl_distribution not in ["fixed", "uniform"]:
            raise ValueError("sl_distribution must be 'uniform' or 'fixed'.")

    def set_val_test_sl(self, sl_test_val):
        self.set_param("sl", sl_test_val, shape=(self.num_SKUs[0],), new=False)



"""Processors for observations can be used to process the input for an agent before it is being passed to the agent."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils/11_obsprocessors.ipynb.

# %% auto 0
__all__ = ['BaseProcessor', 'FlattenTimeDimNumpy', 'ConvertDictSpace', 'AddParamsToFeaturesLEGACY', 'AddParamsToFeatures']

# %% ../nbs/00_utils/11_obsprocessors.ipynb 3
from typing import Union, Optional, List, Tuple, Dict

import numpy as np
from .utils import Parameter, check_parameter_types

import torch
import torch.nn as nn
import torch.nn.functional as F

# %% ../nbs/00_utils/11_obsprocessors.ipynb 4
class BaseProcessor():

    def determine_output_shape(self,
        sample_input: Dict, # sample input
        ) -> Tuple | List:

        """
        Determine the output shape based on the input dictionary.
        """

        output = self.__call__(sample_input)

        if isinstance(output, list):
            return [output_element.shape for output_element in output]
        else:
            return output.shape

# %% ../nbs/00_utils/11_obsprocessors.ipynb 5
class FlattenTimeDimNumpy():

    """
    Preprocessor to flatten the time and feature dimension of the input.
    Used, e.g., to convert time-series data for models that cannot process
    a time dimension such as MLPs or Regression models.
    """

    def __init__(self,
        allow_2d: Optional[bool] = False, #
        batch_dim_included: Optional[bool] = True
        ):
        self.allow_2d = allow_2d
        self.batch_dim_included = batch_dim_included

    def check_input(self,
            input: np.ndarray #
            ):
        """
        Check that the input is a Numpy array with the correct shape.
        """
        # Check if the input is a Numpy array
        if not isinstance(input, np.ndarray):
            raise TypeError(f"Expected input to be a numpy array, but got {type(input)} instead.")

        # Determine expected dimensions based on batch_dim_included
        expected_ndim = 3 if self.batch_dim_included else 2
        allow_ndim = 2 if self.batch_dim_included else 1

        # Check if the input array has the correct dimensions
        if input.ndim == expected_ndim:
            # If the input is 3D, it is valid regardless of allow_2d
            return
        elif input.ndim == allow_ndim:
            # If the input has fewer dimensions, check if the reduced dimension is allowed
            if not self.allow_2d:
                raise ValueError(
                    f"Expected input to have {expected_ndim} dimensions with shape "
                    f"{'(batch_size, timesteps, features)' if self.batch_dim_included else '(timesteps, features)'}, "
                    f"but got shape {input.shape} instead. "
                    f"{allow_ndim}D inputs are not allowed when allow_2d is False."
                )
        else:
            # If the input has an unexpected number of dimensions
            expected_shape_msg = (
                f"Expected input to have {expected_ndim} dimensions with shape "
                f"{'(batch_size, timesteps, features)' if self.batch_dim_included else '(timesteps, features)'}"
                if not self.allow_2d
                else f"Expected input to have {expected_ndim} dimensions with shape "
                     f"{'(batch_size, timesteps, features)' if self.batch_dim_included else '(timesteps, features)'} "
                     f"or {allow_ndim} dimensions with shape "
                     f"{'(batch_size, features)' if self.batch_dim_included else '(features)'}"
            )
            raise ValueError(f"{expected_shape_msg}, but got shape {input.shape} instead.")
            
    def __call__(self,
                input: np.ndarray
                ) -> np.ndarray:
                
        """
        Process the input array by keeping the batch dimension and flattening
        the time and feature dimensions.
        """

        # Validate the input tensor
        self.check_input(input)

        if self.batch_dim_included:
            # If batch dimension is included
            if input.ndim == 2:
                output = input
            else:
                # Keep the batch dimension, flatten time and feature dimensions
                batch_size, timesteps, features = input.shape
                output = input.reshape(batch_size, -1)
        else:
            # If batch dimension is not included
            if input.ndim == 1:
                output = input
            else:
                # Flatten time and feature dimensions
                timesteps, features = input.shape
                output = input.reshape(-1)

        return output

# %% ../nbs/00_utils/11_obsprocessors.ipynb 9
class ConvertDictSpace(BaseProcessor):

    """  

    A utility class to process a dictionary of numpy arrays, with options to preserve or flatten the time dimension.

    Note, this class is only used to preprocess output from the environment without batch dimension.
    
    """

    def __init__(self,
        keep_time_dim: Optional[bool] = False, #If time timension should be flattened as well.
        hybrid_space_params: Optional[Dict] = None, # dict with keys "time" that is a list of observation keys that should keep the time dimension.
        ):
        self.keep_time_dim = keep_time_dim
        self.hybrid_space_params = hybrid_space_params

        if not keep_time_dim and hybrid_space_params is not None:
            raise ValueError("If keep_time_dim is False, hybrid_space_params must be None.")
        if hybrid_space_params is not None and not isinstance(hybrid_space_params, dict):
            raise ValueError("hybrid_space_params must be a dictionary if provided.")


    def __call__(self, 
                input: Dict, # Observation as dict of with numpy arrays
                flatten: bool = True, # whether to flatten composite spaces (non-composite spaces will depend on self.keep_time_dim)
                ) -> List[np.ndarray] | np.ndarray: 

        """
        Process the input dictionary by converting it to a numpy array.
        """

        if self.hybrid_space_params is not None:
            obs_2d = [] # time X features
            obs_1d = [] # features
        else:
            obs = [] # features or time X features

        for counter, (key, value) in enumerate(input.items()):
            if not isinstance(value, np.ndarray):
                raise TypeError(f"Expected input to be a dictionary of numpy arrays, but got {type(value)} instead.")

            if self.hybrid_space_params is not None:
                if key in self.hybrid_space_params["time_series_input"]:
                    obs_2d.append(value)
                    if counter != 0:
                        assert obs_2d[counter].shape[0] == obs_2d[counter-1].shape[0], "All time dimensions must be the same."
                else:
                    obs_1d.append(value.flatten())
            else:
                if self.keep_time_dim:
                    obs.append(value)
                    if counter != 0:
                        assert obs[counter].shape[0] == obs[counter-1].shape[0], "All time dimensions must be the same."
                else:
                    obs.append(value.flatten())

        if self.hybrid_space_params is not None:
            obs_2d = np.concatenate(obs_2d, axis=0)
            obs_1d = np.concatenate(obs_1d, axis=0)
            if flatten:
                return np.concatenate([obs_2d.flatten(), obs_1d], axis=0)
            else:
                return [obs_2d, obs_1d]
        else:
            if obs[0].ndim == 1:
                return np.concatenate(obs, axis=0)
            else:
                return np.concatenate(obs, axis=1)

            return np.concatenate(obs, axis=0)

# %% ../nbs/00_utils/11_obsprocessors.ipynb 10
class AddParamsToFeaturesLEGACY(BaseProcessor):

    """  

    A utility class to process a dictionary of numpy arrays, with options to preserve or flatten the time dimension.
    # TODO: Currently is mixes too many cases like batched input, hybrid input etc. Seperate into more specific obsprocessors.

    """

    def __init__(self,
        environment: object, # The environment object, needed to check if val or train mode,
        keep_time_dim: Optional[bool] = False, #If time timension should be flattened as well.
        hybrid: Optional[bool] = False, # If the param dim should be added as separate vector or concatenated to the features.
        receive_batch_dim: Optional[bool] = False, # If the input contains a batch dimension.
        ):

        self.environment = environment
        self.keep_time_dim = keep_time_dim
        self.hybrid = hybrid
        self.receive_batch_dim = receive_batch_dim

        if not keep_time_dim and hybrid:
            raise ValueError("For flattened vector, hybrid should be be merged with features directy.")


    def __call__(self, 
                input: Dict, # Observation as dict of with numpy arrays
                flatten: bool = False, # whether to flatten composite spaces (non-composite spaces will depend on self.keep_time_dim)
                ) -> List[np.ndarray] | np.ndarray: 

        """
        Process the input dictionary by converting it to a numpy array.
        """

        input = input.copy()
        if self.receive_batch_dim:
            features = input["features"]
            print(features.shape)
            if self.environment.mode == "train":
                features = np.expand_dims(features, axis=0)
                
            if not self.keep_time_dim:
                batch_size, time_steps, feature_dims = features.shape
                new_shape = (batch_size, time_steps*feature_dims)
                features = features.reshape(new_shape)
        else:
            features = input["features"] if self.keep_time_dim else input["features"].flatten()
        del input["features"]

        if self.hybrid:
            if receive_batch_dim:
                raise NotImplementedError("Hybrid not implemented yet for batched input.")
            else:
                obs_1d = [] # features or time X features
                obs_2d = [] # time X features
                obs.append(input["features"])
        
        for counter, (key, value) in enumerate(input.items()):
            if not isinstance(value, np.ndarray):
                raise TypeError(f"Expected input to be a dictionary of numpy arrays, but got {type(value)} instead.")
            
            if value.ndim == 1:
                if features.ndim == 1:
                    features = np.concatenate([features, value])
                else:
                    if self.hybrid:
                        raise NotImplementedError("Hybrid not implemented yet.")
                    # expand value to 2d by copy time dimension
                    else:
                        if self.receive_batch_dim:

                            if features.ndim == 3: # then it is (batch x time x features)
                                value = np.expand_dims(value, axis=0) # add batch dimension
                                value = np.expand_dims(value, axis=1) # add time dimension

                                value = np.repeat(value, features.shape[1], axis=1) # repeat for all time steps
                        
                            else:
                                value = np.expand_dims(value, axis=0) # add batch dimension

                            features = np.concatenate([features, value], axis=-1) # concatenate along feature dimension

                        else:
                            value = np.expand_dims(value, axis=0)
                            value = np.repeat(value, features.shape[0], axis=0)

                            features = np.concatenate([features, value.flatten()])
            
            else:
                if value.shape == features.shape:
                    features = np.concatenate([features, value.flatten()])
                else:
                    raise ValueError(f"Expected input to have the same shape as features, but got {value.shape} instead (feature shape: {features.shape}).")
        
        if self.environment.mode == "train":
            if len(features.shape) == 3:
                features = np.squeeze(features, axis=0) # remove batch dimension
                
        if self.hybrid:
            if flatten:
                raise NotImplementedError("Hybrid not implemented yet.")
            else:
                raise NotImplementedError("Hybrid not implemented yet.")
        else:
            return features
            

# %% ../nbs/00_utils/11_obsprocessors.ipynb 11
class AddParamsToFeatures(BaseProcessor):

    """

    A utility class to process a dictionary of numpy arrays (from dict space), with options to preserve or flatten the time dimension.
    It always adds the parameters to the appropriate dimension. For composite spaces (partially time-series, partially not), use the
    separate AddParamsToFeaturesComposite class.
    
    """

    def __init__(self,
        environment: object, # The environment object, needed to check if val or train mode,
        keep_time_dim: Optional[bool] = False, #If time timension should be flattened as well.
        receive_batch_dim: Optional[bool] = False, # If the input contains a batch dimension.
        ):

        self.environment = environment
        self.keep_time_dim = keep_time_dim
        self.receive_batch_dim = receive_batch_dim

    def __call__(self, 
                input: Dict, # Observation as dict of with numpy arrays
                ) -> List[np.ndarray] | np.ndarray: 

        """
        Process the input dictionary by converting it to a numpy array.
        """

        input = input.copy()
        if self.receive_batch_dim:
            features = input["features"]
                
            if not self.keep_time_dim:
                batch_size, time_steps, feature_dims = features.shape
                new_shape = (batch_size, time_steps*feature_dims)
                features = features.reshape(new_shape)
        else:
            features = input["features"] if self.keep_time_dim else input["features"].flatten()
        del input["features"]
        
        for counter, (key, value) in enumerate(input.items()):
            if not isinstance(value, np.ndarray):
                raise TypeError(f"Expected input to be a dictionary of numpy arrays, but got {type(value)} instead.")
            
            value_shape = value.shape

            if value.ndim == 1:
                if features.ndim == 1:
                    features = np.concatenate([features, value])
                else:
                    if self.receive_batch_dim:

                        if features.ndim == 3: # then it is (batch x time x features)
                            value = np.expand_dims(value, axis=-1) # add value dim (first is batch dim)
                            value = np.expand_dims(value, axis=1) # add time dimension

                            # TODO: check if it always should expand the features dimension
                            if value.shape[0] == 1 and features.shape[0] > 1:
                                value = np.repeat(value, features.shape[0], axis=0)
                            value = np.repeat(value, features.shape[1], axis=1) # repeat for all time steps
                    
                        else:
                            value = np.expand_dims(value, axis=-1) # add value dim (first is batch dim)
                            # TODO: check if it always should expand the features dimension
                            if value.shape[0] == 1 and features.shape[0] > 1:
                                value = np.repeat(value, features.shape[0], axis=0) # repeat for batch

                        features = np.concatenate([features, value], axis=-1) # concatenate along feature dimension

                    else:

                        if features.ndim == 2: # then it is (time x features)
                            value = np.expand_dims(value, axis=0)
                            value = np.repeat(value, features.shape[0], axis=0)

                        features = np.concatenate([features, value.flatten()], axis=-1)
            
            else:
                if value.shape == features.shape:
                    features = np.concatenate([features, value.flatten()])
                else:
                    raise ValueError(f"Expected input to have the same shape as features, but got {value.shape} instead (feature shape: {features.shape}).")

        return features
            



"""Base environment class based on Gymnasium"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/20_base_env/10_base_env.ipynb.

# %% auto 0
__all__ = ['BaseEnvironment']

# %% ../../nbs/20_base_env/10_base_env.ipynb 4
import gymnasium as gym
from abc import ABC, abstractmethod
from typing import Union, List
import numpy as np

from ..utils import MDPInfo, Parameter, set_param
import time

# %% ../../nbs/20_base_env/10_base_env.ipynb 5
class BaseEnvironment(gym.Env, ABC):

    """

    Base class for environments enforcing a common interface.
    """

    def __init__(self,
                    mdp_info: MDPInfo, # MDPInfo object to ensure compatibility with the agents
                    postprocessors: list[object] | None = None,  # default is empty list
                    mode: str = "train", # Initial mode (train, val, test) of the environment
                    return_truncation: str = True, # whether to return a truncated condition in step function
                    horizon_train: int | str = "use_all_data" # horizon of the training data
                    ) -> None: #

        super().__init__()

        self.horizon_train = horizon_train

        self.return_truncation = return_truncation

        self._mode = mode
        self._mdp_info = mdp_info
        
        if mode == "train": 
            self.train()
        elif mode == "val":
            self.val()
        elif mode == "test":
            self.test()
        else:
            raise ValueError("mode must be 'train', 'val', or 'test'")

        self.postprocessors = postprocessors or []

    def set_param(self,
                        name: str, # name of the parameter (will become the attribute name)
                        input: Parameter | int | float | np.ndarray | List | None, # input value of the parameter
                        shape: tuple = (1,), # shape of the parameter
                        new: bool = False # whether to create a new parameter or update an existing one
                        ) -> None: #
        
        """
        Set a parameter for the environment. It converts scalar values to numpy arrays and ensures that
        environment parameters are either of the Parameter class of Numpy arrays. If new is set to True, 
        the function will create a new parameter or update an existing one otherwise. If new is set to
        False, the function will raise an error if the parameter does not exist.
        """

        set_param(self, name, input, shape, new)

    def return_truncation_handler(self, observation, reward, terminated, truncated, info):
        """ 
        Handle the return_truncation attribute of the environment. This function is called by the step function

        """

        if self.return_truncation:
            return observation, reward, terminated, truncated, info
        else:
            return observation, reward, terminated, info

    def step(self, action):
        
        """
        Step function of the environment. Do not overwrite this function. 
        Instead, write the step_ function. Note that the postprocessor is applied here.

        """
        
        ## apply postprocessor
        for postprocessor in self.postprocessors:
            action = postprocessor(action)

        observation, reward, terminated, truncated, info = self.step_(action)

        return self.return_truncation_handler(observation, reward, terminated, truncated, info)
    
    def add_postprocessor(self, postprocessor: object): # post-processor object that can be called via the "__call__" method
        """Add a postprocessor to the agent"""
        self.postprocessors.append(postprocessor)

    @staticmethod
    def step_(self, action):
        """
        Step function of the environment. It is a wrapper around the step function that handles the return_truncation
        attribute of the environment. It must return the following: observation, reward, terminated, truncated, info

        """
        pass


    @property
    def mdp_info(self):
        """
        Returns: The MDPInfo object of the environment.

        """
        return self._mdp_info

    @property
    def info(self):
        """
        Returns: Alternative call to the method for mushroom_rl.

        """
        return self.mdp_info

    @property
    def mode(self):
        """
        Returns: A string with the current mode (train, test val) of the environment.

        """
        return self._mode

    @abstractmethod
    def set_action_space(self):
        """
        Set the action space of the environment.

        """
        pass

    @abstractmethod
    def set_observation_space(self):
        """
        Set the observation space of the environment.
        In general, this can be also a dict space, but the agent must have the appropriate pre-processor.

        """
        pass

    @abstractmethod
    def get_observation(self):
        """
        Return the current observation. Typically constructed from the output of the dataloader and 
        internal dynamics (such as inventory levels, pipeline vectors, etc.) of the environment.

        """
        pass

    @abstractmethod
    def reset(self):
        """
        Reset the environment. This function must be provided, using the function self.reset_index() to 
        handle indexing. It needs to account for the current training mode train, val, or test and handle
        the horizon_train param. See the reset function for the NewsvendorEnv for an example.

        """
        pass
    
    def set_index(self, index=None):
        
        """
        Handle the index of the environment.

        """

        if index is not None:
            self.index = index
        else:
            self.index += 1
        
        truncated = True if self.index >= self.max_index_episode else False
        
        return truncated

    def get_start_index(self,
        start_index: int | str = None, # index to start from
        ) -> int:
        
        """ Determine if the start index is random or 0,
        depending on the state of the environment and training
        process (over entire train set or in shorter episodes) """

        if start_index is None:
            if self._mode == "train":
                if self.horizon_train == "use_all_data":
                    start_index = 0
                elif hasattr(self.dataloader, "is_distribution") and self.dataloader.is_distribution:
                    start_index = 0
                else:
                    start_index = "random"
            elif self._mode == "val":
                start_index = 0
            elif self._mode == "test":
                start_index = 0
            else:
                raise ValueError("Mode not recognized.")
            
        else:
            start_index = start_index
        
        return start_index

    def reset_index(self,
        start_index: Union[int,str]):

        """

        Reset the index of the environment. If start_index is an integer, the index is set to this value. If start_index is "random",
        the index is set to a random integer between 0 and the length of the training data.

        """

        start_index = self.get_start_index(start_index)
 
        if start_index=="random":
            if self.mode == "train":
                if self.dataloader.len_train is not None and self.dataloader.len_train > self.mdp_info.horizon:
                    random_index = np.random.choice(self.dataloader.len_train-self.mdp_info.horizon)
                else:
                    random_index = 0
                self.start_index = random_index 
            else:
                raise ValueError("start_index cannot be 'random' in val or test mode")
        elif isinstance(start_index, int):
            self.start_index = start_index
        else:
            raise ValueError("start_index must be an integer or 'random'")

        self.max_index = self.dataloader.len_train if self.mode == "train" else self.dataloader.len_val if self.mode == "val" else self.dataloader.len_test
        self.max_index -= 1
        self.max_index_episode = np.minimum(self.max_index, self.start_index+self.mdp_info.horizon)
        if self.mode == "test" or self.mode == "val":
            self.max_index_episode += 1
    
        truncated = self.set_index(self.start_index) # assuming we only start randomly during training.
        
        return truncated

    def update_mdp_info(self, gamma=None, horizon=None):
        
        """
        Update the MDP info of the environment.

        """
        if gamma is not None:
            self._mdp_info.gamma = gamma
        if horizon is not None:
            self._mdp_info.horizon = horizon

    def train(self, update_mdp_info=True):
        """
        Set the environment in training mode by both setting the internal state self._train and the dataloader. 
        If the horizon is set to "use_all_data", the horizon is set to the length of the training data, otherwise
        it is set to the horizon_train attribute of the environment. Finally, the function updates the MDP info
        and resets with the new state.

        """
        self._mode = "train"

        if hasattr(self, "dataloader"):
            self.dataloader.train()

            if hasattr(self, "horizon_train"):
                if self.horizon_train == "use_all_data":
                    horizon = self.dataloader.len_train
                else:
                    horizon = self.horizon_train
        else:
            horizon = self.mdp_info.horizon

        if update_mdp_info:
            self.update_mdp_info(gamma=self.mdp_info.gamma, horizon=horizon)

        self.reset()
    
    def val(self, update_mdp_info=True):
        """
        Set the environment in validation mode by both setting the internal state self._val and the dataloader.
        The horizon of val is always set to the length of the validation data. Finally, the function updates the MDP info
        and resets with the new state.

        """
        self._mode = "val"

        if hasattr(self, "dataloader"):
            self.dataloader.val()
            horizon = self.dataloader.len_val
        else:
            horizon = self.mdp_info.horizon

        if update_mdp_info:
            self.update_mdp_info(gamma=self.mdp_info.gamma, horizon=horizon)

        self.reset()

    def test(self, update_mdp_info=True):
        """
        Set the environment in testing mode by both setting the internal state self._test and the dataloader.
        The horizon of test is always set to the length of the test data. Finally, the function updates the MDP info
        and resets with the new state.

        """
        self._mode = "test"

        if hasattr(self, "dataloader"):
            self.dataloader.test()
            horizon = self.dataloader.len_test
        else:
            horizon = self.mdp_info.horizon

        if update_mdp_info:
            self.update_mdp_info(gamma=self.mdp_info.gamma, horizon=horizon)

        self.reset()

    def set_return_truncation(self, return_truncation: bool): # whether or not to return the truncated condition in the step function
        
        """
        Set the return_truncation attribute of the environment.
        """

        self.return_truncation = return_truncation

    def stop(self):
        """
        Stop the environment. This function is used to ensure compatibility with the Core of mushroom_rl.

        """
        pass




"""Dynamic inventory management problem with inventory carry-over. Can be used to model the Lost Sales problem (when fixed cost are set to 0), and the Multi-Period Fixed Cost problem (when fixed cost are larger than 0)."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/21_envs_inventory/30_multi_period_envs.ipynb.

# %% auto 0
__all__ = ['MultiPeriodEnv']

# %% ../../../nbs/21_envs_inventory/30_multi_period_envs.ipynb 3
from abc import ABC, abstractmethod
from typing import Union, Tuple

from ...utils import Parameter, MDPInfo, check_parameter_types
from ...dataloaders.base import BaseDataLoader
from .base import BaseInventoryEnv
from .inventory_utils import OrderPipeline

import gymnasium as gym

import numpy as np
import time

# %% ../../../nbs/21_envs_inventory/30_multi_period_envs.ipynb 4
class MultiPeriodEnv(BaseInventoryEnv, ABC):
    
    """
    XXX
    """

    def __init__(self,
        
        underage_cost: np.ndarray | Parameter | int | float = 1,  # underage cost per unit
        overage_cost: np.ndarray | Parameter | int | float = 0,  # overage cost per unit (zero in most cases)

        fixed_ordering_cost: np.ndarray | Parameter | int | float = 0,  # fixed ordering cost (applies per SKU, not jointly)
        variable_ordering_cost: np.ndarray | Parameter | int | float = 0,  # variable ordering cost per unit
        holding_cost: np.ndarray | Parameter | int | float = 1,  # holding cost per unit

        start_inventory: np.ndarray | Parameter | int | float = 0,  # initial inventory
        max_inventory: np.ndarray | Parameter | int | float = np.inf,  # maximum inventory

        inventory_pipeline_params: dict | None = None,  # parameters for the inventory pipeline, only lead_time_mean must be given. 

        q_bound_low: np.ndarray | Parameter | int | float = 0,  # lower bound of the order quantity
        q_bound_high: np.ndarray | Parameter | int | float = np.inf,  # upper bound of the order quantity
        dataloader: BaseDataLoader = None,  # dataloader
        num_SKUs: int | None = None,  # if None, it will be inferred from the DataLoader
        gamma: float = 1,  # discount factor
        horizon_train: int | str = 100,  # if "use_all_data", then horizon is inferred from the DataLoader
        postprocessors: list[object] | None = None,  # default is an empty list
        mode: str = "train",  # Initial mode (train, val, test) of the environment
        return_truncation: bool = True,  # whether to return a truncated condition in step function
        step_info_verbosity = 0,  # 0: no info, 1: some info, 2: all info

    ) -> None:

        self.print=False

        num_SKUs = dataloader.num_units if num_SKUs is None else num_SKUs
        if not isinstance(num_SKUs, int):
            raise ValueError("num_SKUs must be an integer.")
        
        self.set_param("num_SKUs", num_SKUs, new=True)

        self.set_param("q_bound_low", q_bound_low, shape=(num_SKUs,), new=True)
        self.set_param("q_bound_high", q_bound_high, shape=(num_SKUs,), new=True)

        self.set_param("fixed_ordering_cost", fixed_ordering_cost, shape=(num_SKUs,), new=True)
        self.set_param("variable_ordering_cost", variable_ordering_cost, shape=(num_SKUs,), new=True)
        self.set_param("holding_cost", holding_cost, shape=(num_SKUs,), new=True)

        self.set_param("start_inventory", start_inventory, shape=(num_SKUs,), new=True)
        self.set_param("max_inventory", max_inventory, shape=(num_SKUs,), new=True)
        self.start_inventory = self.start_inventory.astype(float)

        inventory_pipeline_params = inventory_pipeline_params or {}
        inventory_pipeline_params["num_units"] = int(self.num_SKUs[0])
        self.order_pipeline = OrderPipeline(**inventory_pipeline_params)
        self.inventory = self.start_inventory.copy()

        self.set_observation_space(dataloader.X_shape)
        self.set_action_space(dataloader.Y_shape, low = self.q_bound_low, high = self.q_bound_high)

        mdp_info = MDPInfo(self.observation_space, self.action_space, gamma=gamma, horizon=horizon_train)

        check_parameter_types(step_info_verbosity, parameter_type=int)
        self.step_info_verbosity = step_info_verbosity
        
        super().__init__(mdp_info=mdp_info,
                            postprocessors = postprocessors, 
                            mode=mode, return_truncation=return_truncation,
                            underage_cost=underage_cost,
                            overage_cost=overage_cost, 
                            dataloader=dataloader,
                            horizon_train = horizon_train)

    def step_(self, 
            action: np.ndarray # order quantity
            ) -> Tuple[np.ndarray, float, bool, bool, dict]:

        """
        XXX.

        """

        # Most agent give by default a batch dimension which is not needed for a single period action.
        # If action shape size is 2 and the first dimensiion is 1, then remove it
        if action.ndim == 2 and action.shape[0] == 1:
            action = np.squeeze(action, axis=0)  # Remove the first dimension

        variable_ordering_cost = action * self.variable_ordering_cost
        fixed_ordering_cost = np.where(action > 0, self.fixed_ordering_cost, 0)

        orders_arriving = self.order_pipeline.step(action) # add orders to pipeline and get arriving orders

        # print("action:", action)
        # print("unit cost:", self.variable_ordering_cost)
        # print("variable_ordering_cost:", variable_ordering_cost)

        # print("unit_fixed_ordering_cost:", self.fixed_ordering_cost)
        # print("fixed_ordering_cost:", fixed_ordering_cost)

        # print("old inventory:", self.inventory)
        # print("orders arriving:", orders_arriving)
        # print("demand:", self.demand)

        self.inventory += orders_arriving
        self.inventory -= self.demand
        self.inventory = np.minimum(self.inventory, self.max_inventory)

        

        # check where the inventory is below 0
        underage_quantity = np.maximum(-self.inventory, 0)
        underage_cost = underage_quantity * self.underage_cost
        # print("underage_quantity:", underage_quantity, "underage_cost:", underage_cost)
        self.inventory = np.maximum(self.inventory, 0)

        holding_cost = self.inventory * self.holding_cost
        # print("holding_cost:", holding_cost)
        # print("new inventory:", self.inventory)

        total_cost_step = variable_ordering_cost + fixed_ordering_cost + underage_cost + holding_cost
        reward = -np.sum(total_cost_step) # negative because we want to minimize the cost

        terminated = False # in this problem there is no termination condition
        
        info = {}
        if self.step_info_verbosity > 1:
            info["demand"] = self.demand.copy()
            info["action"] = action.copy()
            info["cost_per_SKU"] = total_cost_step.copy()
        if self.step_info_verbosity > 0:
            info["variable_ordering_cost"] = variable_ordering_cost.copy()
            info["fixed_ordering_cost"] = fixed_ordering_cost.copy()
            info["underage_cost"] = underage_cost.copy()
            info["holding_cost"] = holding_cost.copy()

        # Set index will set the index and return True if the index is out of bounds
        truncated = self.set_index()

        if truncated:

            # observation = self.observation_space.sample()
            # for key, value in observation.items():
            #     observation[key] = np.zeros_like(value)
            # demand = np.zeros_like(self.action_space.sample())

            observation, self.demand = self.get_observation()

            return observation, reward, terminated, truncated, info
        
        else:

            observation, self.demand = self.get_observation()

            if self.print:
                print("next_period:", self.index+1)
                print("next observation:", observation)
                print("next demand:", self.demand)
                time.sleep(3)

            return observation, reward, terminated, truncated, info
        
    def get_observation(self):
        
        """
        Return the current observation. This function is for the simple case where the observation
        is only an x,y pair. For more complex observations, this function should be overwritten.

        """

        
        X_item, Y_item = self.dataloader[self.index]

        observation = {
            "features": X_item,
            "order_pipeline": self.order_pipeline.get_pipeline(),
            "inventory:": self.inventory,
        }

        return observation, Y_item

        
    def reset(self,
        start_index: int | str = None, # index to start from
        state: np.ndarray = None # initial state
        ) -> Tuple[np.ndarray, bool]:

        """
        Reset function for the Newsvendor problem. It will return the first observation and demand.
        For val and test modes, it will by default reset to 0, while for the train mode it depends
        on the paramter "horizon_train" whether a random point in the training data is selected or 0
        """

        self.order_pipeline.reset()
        self.inventory = self.start_inventory.copy()

        truncated = self.reset_index(start_index)

        observation, self.demand = self.get_observation()

        # print("in reset:", observation)
        
        return observation

    def set_observation_space(self,
                            feature_shape: tuple, # shape of the dataloader features
                            feature_low: Union[np.ndarray, float] = -np.inf, # lower bound of the observation space
                            feature_high: Union[np.ndarray, float] = np.inf, # upper bound of the observation space
                            samples_dim_included = True # whether the first dimension of the shape input is the number of samples
                            ) -> None:
        
        '''
        Set the observation space of the environment.

        '''

        # To handle cases when no external information is available (e.g., parametric NV)

        spaces = {}

        if isinstance(feature_shape, tuple):
            if samples_dim_included:
                feature_shape = feature_shape[1:] # assumed that the first dimension is the number of samples
            spaces["features"] = gym.spaces.Box(low=feature_low, high=feature_high, shape=feature_shape, dtype=np.float32)
        elif feature_shape is None:
            pass
        else:
            raise ValueError("Shape for features must be a tuple or None")

        len_pipeline, num_products = self.order_pipeline.shape
        order_pipeline_shape = self.order_pipeline.shape
        
        # add dim with lengh of order pipeline and copy values in that dimension

        q_bound_low_full = np.tile(self.q_bound_low, (len_pipeline, 1))
        q_bound_high_full = np.tile(self.q_bound_high, (len_pipeline, 1))
        
        spaces["order_pipeline"] = gym.spaces.Box(low=q_bound_low_full, high=q_bound_high_full, shape=order_pipeline_shape, dtype=np.float32)
        spaces["inventory"] = gym.spaces.Box(low=0, high=self.max_inventory, shape=(int(self.num_SKUs[0]),), dtype=np.float32)
        
        self.observation_space = gym.spaces.Dict(spaces)



"""PPO based agent"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/51_RL_agents/10_PPO_agents.ipynb.

# %% auto 0
__all__ = ['PPOAgent']

# %% ../../../nbs/51_RL_agents/10_PPO_agents.ipynb 4
import logging

# set logging level to INFO
logging.basicConfig(level=logging.INFO)

from abc import ABC, abstractmethod
from typing import Union, Optional, List, Tuple
import numpy as np
import os

from ...envs.base import BaseEnvironment
from .mushroom_rl import MushroomBaseAgent
from ...utils import MDPInfo, Parameter
from ...obsprocessors import FlattenTimeDimNumpy
from ...RL_approximators import MLPState, MLPActor
from ...postprocessors import ClipAction

from ...dataloaders.base import BaseDataLoader

from mushroom_rl.algorithms.actor_critic.deep_actor_critic import PPO
from mushroom_rl.policy import GaussianTorchPolicy

import torch
import torch.optim as optim
import torch.nn.functional as F
from torchinfo import summary

import time

# %% ../../../nbs/51_RL_agents/10_PPO_agents.ipynb 5
class PPOAgent(MushroomBaseAgent):

    """
    XXX
    """

    # TODO: Make structure same to SAC with TD3 base class

    def __init__(self, 
                environment_info: MDPInfo,

                learning_rate_actor: float = 3e-4,
                learning_rate_critic: float | None = None, # If none, then it is set to learning_rate_actor
                batch_size: int = 64,
                hidden_layers: List = None, # if None, then default is [64, 64]
                activation: str = "relu", # "relu", "sigmoid", "tanh", "leakyrelu", "elu"
                # tau: float = 0.005,
                std_0: float = 0.1,
                n_epochs_policy: int = 4,
                eps_ppo: float = 0.2,
                lam: float = 0.95,
                ent_coeff: float = 0.,
                n_steps_per_fit=1000,

                drop_prob: float = 0.0,
                batch_norm: bool = False,
                init_method: str = "xavier_uniform", # "xavier_uniform", "xavier_normal", "he_normal", "he_uniform", "normal", "uniform"

                optimizer: str = "Adam", # "Adam" or "SGD" or "RMSprop"  
                loss: str = "MSE", # currently only MSE is supported     
                obsprocessors: list | None = None,      # default: []
                device: str = "cpu", # "cuda" or "cpu"
                agent_name: str | None = "SAC",
                ):

        self.n_steps_per_fit=n_steps_per_fit

        # The standard TD3 agent needs a 2D input, so we need to flatten the time dimension
        flatten_time_dim_processor = FlattenTimeDimNumpy(allow_2d=True, batch_dim_included=False)
        obsprocessors = (obsprocessors or []) + [flatten_time_dim_processor]

        use_cuda = self.set_device(device)

        hidden_layers = hidden_layers or [64, 64]

        OptimizerClass=self.get_optimizer_class(optimizer)
        learning_rate_critic = learning_rate_critic or learning_rate_actor
        lossfunction = self.get_loss_function(loss)

        input_shape = self.get_input_shape(environment_info.observation_space)
        actor_output_shape = environment_info.action_space.shape

        input_shape = self.convert_recursively_to_int(input_shape)
        actor_output_shape = self.convert_recursively_to_int(actor_output_shape)

        policy_params = dict(network=MLPActor,

                                input_shape=input_shape,
                                output_shape=actor_output_shape,

                                hidden_layers=hidden_layers,
                                activation=activation,
                                drop_prob=drop_prob,
                                batch_norm=batch_norm,
                                init_method=init_method,

                                use_cuda=use_cuda,
                                dropout=self.dropout,

                                st_0=std_0,

                                )
                            
        policy = GaussianTorchPolicy(**policy_params)

        actor_optimizer = {'class': OptimizerClass,
            'params': {'lr': learning_rate_actor}} 

        critic_params = dict(network=MLPState,
                optimizer={'class': OptimizerClass,
                        'params': {'lr': learning_rate_critic}}, 
                loss=lossfunction,
                input_shape=input_shape,
                output_shape=(1,),

                hidden_layers=hidden_layers,
                activation=activation,
                drop_prob=drop_prob,
                batch_norm=batch_norm,
                init_method=init_method,

                use_cuda=use_cuda,
                dropout=self.dropout,)

        self.agent = PPO(
            mdp_info=environment_info,
            policy=policy,
            actor_optimizer=actor_optimizer,
            critic_params=critic_params,
            n_epochs_policy=n_epochs_policy,
            batch_size=batch_size,
            eps_ppo=eps_ppo,
            lam=lam,
            ent_coeff=ent_coeff,
            critic_fit_params=None
        )

        super().__init__(
            environment_info=environment_info,
            obsprocessors=obsprocessors,
            device=device,
            agent_name=agent_name
        )

        logging.info("Actor network:")
        if logging.getLogger().isEnabledFor(logging.INFO):
            input_size = self.add_batch_dimension_for_shape(input_shape)
            print(summary(self.actor, input_size=input_size))
            time.sleep(.2)

        logging.info("Critic network:")
        if logging.getLogger().isEnabledFor(logging.INFO):
            input_size = self.add_batch_dimension_for_shape(input_shape)
            print(summary(self.critic, input_size=input_size))

    def get_network_list(self, set_actor_critic_attributes: bool = True):
        """ Get the list of networks in the agent for the save and load functions
        Get the actor for the predict function in eval mode """

        critic = self.agent._V._impl.model.network
        actor = self.agent.policy._mu._impl.model.network

        networks = []
        networks.append(critic)
        networks.append(actor)

        if set_actor_critic_attributes:
            return networks, actor, critic
        else:
            return networks
